#!/usr/bin/env perl
# make-names.pl

use warnings;
use strict;
BEGIN {
  use File::Basename;
  unshift @INC, dirname(readlink($0) || $0); 
  require 'cdgsetup.pl';
}
use CDG;
use CDG::Utils qw(:functions);

# how many forms we produced
our $forms = 0;

use Getopt::Std;
our $opt_c;
getopts('c:');

# -c gives a corpus of required forms. If a corpus is given,
# only forms in the corpus are generated, to save
# space in the grammar file.
our %forms;
if ($opt_c) {

  sub deutsch_lc {
    my $x = shift;
    $x = lc $x;
    $x =~ tr///;
    return $x;
  }

  sub deutsch_uc {
    my $x = shift;
    $x = uc $x;
    $x =~ tr///;
    return $x;
  }

  # disable `INFO: Loading' message
  commandEval 'set verbosity 0';

  for (split(/,/,"$opt_c")) {
    commandEval "load $_";
  }

  my $all = InputStruct_lattices_get($CDG::inputCurrentGrammar);
  for (my $l = $all; $l; $l = listNext $l) {
    my $lat = listElement($l);
    for (my $m = LatticeStruct_arcs_get $lat; $m; $m = listNext $m) {
      my $w = ArcStruct_word_get (listElement $m);

      $forms{$w}++;

      if ($w =~ s/^.+[-\/]//) {
        $forms{$w}++;
      }

      # make entry for 'Bericht' when encountering 'BERICHT'
      if ($w =~ /^([A-Z])([A-Z]+)$/) {
        my ($x,$y) = ($1, $2);
        $y = deutsch_lc $y;
        $forms{"$x$y"}++;
      }

      # make entry for 'Forum' when encountering 'Diskussionsforum'
      if ($w =~ /....../) {
        my $x = $w;
        while ($x =~ s/^.//) {
          if ($x =~ /^(.)(.{2,})$/) {
            my ($a,$b) = ($1,$2);
            $a = deutsch_uc $a;
            $forms{"$a$b"}++;
          }
        }
      }

      my $l = deutsch_lc $w;
      $forms{$l}++;
    }
  }
}

# should we print this form?
# FALSE if there is a whitelist of forms and this one isn't on it.
# TRUE otherwise.
sub wanted {
  my $form = shift;
  return (!%forms || exists($forms{$form}));
}

sub print_identifier {
  my $x = shift;
  $x =~ s/_/ /g;
  $x =~ s/([\'\"])/\\$1/g;
  if($x =~ /[^a-zA-Z]/) {
    $x = "\"$x\"";
  }
  print $x;
}

sub print_name {
  my $type = shift;
  my $name = shift;
  my $case = shift;
  my $number = shift;
  my $gender = shift;

  return unless wanted $name;
  $forms++;

  print_identifier $name;

  my $sib = ($name =~ /(|s|ch|c)$/) ? ',sibilant:yes' : '';

  print " := [cat:NE,subcat:$type,
  case:$case,person:third,gender:$gender,number:$number,sort:top $sib];\n";

}

our %done;
sub make_name {

  my $apostrophe = 0;
  my $fixed = 0;
  my $plural = 0;
  my $gender = 'bot';

  my $type = shift;
  my $line = shift;

  $line =~ /^([^ ]+)(.*)$/;
  my ($name, $rest) = ($1,$2);
  $name =~ s/_/ /g;
  $apostrophe = 1 if $name =~ /(|s|ch|c|x|z)$/;

  # check for duplicates
  if ($done{"$type/$name"}++) {
    warn "Duplicate definition for $type `$name'!\n";
    return;
  }

  for(split(/ +/,$rest)) {
    next if /^$/;

    if($_ eq "fix") {
      $fixed = 1;
    } elsif($_ eq "plu") {
      $plural = 1;
    } elsif($_ eq "fem") {
      $gender = 'fem';
    } elsif($_ eq "masc") {
      $gender = 'masc';
    } else {
      die "Bad line: $line\n";
    }
  }

  # singular forms

  # non-inflectable names (ETH)
  if($fixed ||
     ($gender eq 'fem' && $type ne 'Vorname')) {
    print_name($type, $name,'bot','sg', $gender);
  }

  # names ending in sibilants (Ross)
  elsif($apostrophe) {
    print_name($type, $name,'(nom|dat|acc)','sg', $gender);
    print_name($type, "${name}'",'gen','sg', $gender);

    # Names in `ch' can form the genitive both with ' and with s:
    # Karl Bartsch --> Karl Bartsch' Werke
    # Frankreich   --> Frankreichs Regierung
    if($name =~ /ch?$/) {
      print_name($type, "${name}s",'gen','sg', $gender);
    }
  }

  # pluralia tantum (USA)
  elsif($plural) {
    print_name($type, $name,'bot','pl', $gender);
  }

  # normal names (Jennifer)
  else {

    # country and similar names are often uninflected even in the
    # Genitive:
    #
    # Lybiens Diktator -- die Wirtschaft des heutigen Lybien
    #
    # Therefore we must provide for both possibilities.
    if($type =~ /Region|Vorname|Nachname/) {
      print_name($type, $name,'bot','sg', $gender);
    } else {
      print_name($type, $name,'(nom|dat|acc)','sg', $gender);
    }
    print_name($type, "${name}s",'gen','sg', $gender);
    print_name($type, "${name}'s",'gen','sg', $gender);
  }

  # plural forms
  if($type eq 'Produkt') {
    print_name($type, "${name}s",'bot','pl', $gender);
  }

  if($type eq 'Nachname') {
    print_name($type, "${name}s",'bot','pl', $gender);
  }

}

while (<>) {
  chomp;

  # find comments
  next if /^\#/;

  # find name declaration
  if (/^(Firma|Nachname|Produkt|Region|Vorname) +(.*)$/) {
    make_name ($1, $2);
  }

}

print STDERR "Generated $forms names.\n";
