#!/usr/bin/env perl
# cdg/grammar/deutsch/make-verbs.pl
#
# This program generates lexical entries and ancillary structures
# (hierarchies, maps and constraints) for modelling verbs in
# the constraint dependency grammar of German.
#
# Input to this program consists in paradigm declarations, specifying
# how a particular German base verb behaves under inflection, and verb
# declarations, specifying the valence and other properties of the
# actual verb (usually a composite of a base verb and one or more
# prefixes). See Verben.txt for the exact input syntax.
#
# The output of this program is twofold: lexicon entries are printed
# to STDOUT, while other structures necessary for handling German
# verbs are written to the file AVZ.cdg. This is so that the lexicon
# entries (which are very numerous) can be auto-loaded by cdg, while
# constraints and hierarchies go into the main grammar.
use warnings;
use strict;
BEGIN {
  use File::Basename;
  unshift @INC, dirname(readlink($0) || $0); 
  require 'cdgsetup.pl';
}
use CDG;
use CDG::Utils qw(:functions);

use Getopt::Std;
our $opt_c;
getopts('c:');

# The command-line option -c specifies a comma-separated list of files
# that are assumed to be the corpus for which we should generate a
# lexicon. If a corpus is specified, all forms that are not present in
# it are not actually generated to save space.
our %forms;

sub deutsch_lc {
  my $x = shift;
  $x = lc $x;
  $x =~ tr///;
  return $x;
}

if ($opt_c) {

  # disable `INFO: Loading' message
  commandEval 'set verbosity 0';

  for (split(/,/,"$opt_c")) {
    commandEval "load $_";
  }

  my $all = InputStruct_lattices_get($CDG::inputCurrentGrammar);
  for (my $l = $all; $l; $l = listNext $l) {
    my $lat = listElement($l);
    for (my $m = LatticeStruct_arcs_get $lat; $m; $m = listNext $m) {
      my $w = ArcStruct_word_get (listElement $m);
      $w =~ s/^.+-//;
      $forms{$w}++;
      my $l = deutsch_lc $w;
      $forms{$l}++;
    }
  }
}

# All paradigm info is stored here. These hashes contain paradigms of
# verb inflection as returned by make_paradigm(), keyed by the
# infinitive form. Two hashes are provided because some verbs are
# inflected both regularly and irregularly (e.g. `hngen').
our %regular_paradigms;
our %irregular_paradigms;

# When particular forms alternate with other forms indistinguishably,
# you can just specify individual alternation as a `variant' without
# having to hack the paradigms.
our %variants;

# The input file may define macros to avoid typing repeated text.
# They are stored here, e.g.: $macros{SEIN} eq 'perfect:sein'.
our %macros;

# Frequencies of all verb frames encountered:
# e.g. %valences{'A+D'} == 12
our %valences;

# Frequencies all argument substitution possibilities:
# e.g. $argtypes{AC} == 132
our %argtypes;

# All labels relevant for verb valence stored by their sigils
our %labels =
    ( a => 'OBJA',
      b => 'OBJA2',
      c => 'OBJC',
      d => 'OBJD',
      e => 'PRED',
      g => 'OBJG',
      i => 'OBJI',
      k => 'KOM',
      p => 'OBJP',
      n => 'PN',
      s => 'S',
      x => 'AUX' );

# Hash of hashes of allowed separable verb-prefix pairs.
#
# For each possible infinitive and separable prefix,
# specifies the proper frame type.
#
# Example: $avz{geben}{aus} == 'A',  $avz{verkaufen}{aus} == 'A',
# but $avz{geben}{ent} should not exist.
our %avz;

# Like %avz, but for requirements similar to valence.
our %objp;
our %objc;
our %subjc;
our %loctype;

# Hash of occurring separable prefixes, e.g. $prefixes{aus} == 1,
# $prefixes{allein} == 1, but $prefixes{ge} should not exist.
#
# This is not merely used for creating statistics; the actual
# [cat:PTKVZ] entries are generated by this script to save useless
# lexicon items.
our %prefixes;

# List of decoded verb declarations, e.g.
#
# $verbs[0] == { base => gehen, reg => 'regular', valence => 'A' ... }
#
# All `verb' declarations are parsed before any output is generated;
# the reason is that we want to avoid generating forms for separable
# verbs if they are also present as ordinary verbs, and we can only
# know that after we have read all input. This array holds the
# declarations between the two passes.
our @verbs;

# Hash of all complete infinitives, e.g. $infinitives{ausgeben} == 1,
# $infinitives{geben} == 1, $infinitives{gelingen} == 1,
# but $infinitives{lingen} should not exist.
#
# We need this to decide whether or not to create the seperated SVO
# forms of separable verbs. For example, if we have a declaration of
# `wandeln' and one of `um<wandeln', then we need not generate the
# form `wandelt [um]' because we've already done it. But if we have
# `ab<flauen' only, we do need to generate `flaut' because the base
# verb does not exist.
#
# In general, we do not want to create the separated SVO version of a
# separable verb if the non-separable verb also exists. Therefore we
# check the hash %infinitives strategically before actually creating
# such a form -- see print_finite_forms().
our %infinitives;

# Holds the infinitives of verbs that are ambiguous with respect to
# the stressing of their prefixes, e.g. "berlegen" (nachdenken) vs.
# "berlegen" (bedecken). These verbs need to specify the feature
# stress so that an annotation can specify which variant it meant.
our %ambiguous_stress;

# how many forms we produced altogether
our $forms = 0;

# Should we print this form?
#
# FALSE if there is a whitelist of forms and this one isn't on it.
# TRUE otherwise.
sub wanted {
  my $form = shift;
  return (!%forms || exists($forms{$form}));
}

# insert a macro definition into the table
sub make_macro ($ $ ) {
  my ($macro,$expansion) = @_;
  $macros{$macro} = $expansion;
}

# expand instances of macros
sub expand_macros {
  my $in = shift;
  my $out = $in;

  while ($in =~ /\b([A-Z]+)\b/g) {
    my $macro = $1;
    if (exists $macros{$macro}) {
      $out =~ s/\b$macro\b/$macros{$macro}/g;
    }
  }

  return $out;
}


# Distinguish full verbs from auxiliary and modal verbs.
#
# The distinction is not as clear-cut as STTS seems to think; even the
# auxiliary verbs have separable variants (`los<sein', `vor<haben'),
# and it is not clear whether `hat' should be tagged as a full verb
# just by virtue of having a `vor' complement. Here we just look at
# the infinitive and return 'A', 'M', or 'V'.
sub classify {
  my $verb = shift;
  if ($verb =~ /^(sein|haben|werden)$/) {
    return 'A'
  } elsif ($verb =~ /^(drfen|wollen|mgen|knnen|sollen|mssen)$/) {
    return 'M'
  } else {
    return 'V'
  }
}

# Decode pre-processed verb declaration.
#
# There are five possible types of verbs.
# $type distinguishes between them as follows:
# `none' - no prefixes (`gehen')
# `stressed' - stressed prefix (`assteigen')
# `unstressed' - unstressed prefix (`verschern')
# `su' - two prefixes, first stressed (`bererfllen')
# `us' - two prefixes, second stressed (`verenbaren')
#
# There may be more (`mtvereinbaren'), but in that case
# we can treat two of them as a unit.
#
# This sub takes the verb declaration as per Verben.txt
# and extracts the verb type and all components.
#
# Example: 'aus<er>whlen' ==> ('su', 'aus', 'er', 'whlen')
#
# The boundaries between a nonseparable prefix and a long base verb
# are unclear - is it `ge>lingen' or `gelingen'? - but we only need to
# make sure that our paradigm declarations and our verb declarations
# match. For this case, the best solution is to declare an irregular
# paradigm `lingen' and the two verbs `ge>lingen' and `mi>lingen'.
sub decode_declaration {
  my $decl = shift;
  my $valence = shift;
  my $e = shift;

  my $objp = $$e{objp} || '';
  my $objc = $$e{objc} || 'top';
  my $loctype = $$e{loctype} || '';
  my $subjc = $$e{nimmt_Subjektsatz} || '';

  sub count_infinitive {
    my $i = shift;
    if(1 < ++$infinitives{$i}) {
#      warn "stress-ambiguous verb: $i!\n";
      $ambiguous_stress{$i} = 1;
    }
  }


  if ($decl =~ /^(.+)<(.+)>(.+)$/) {
    $avz{$2.$3}{$1} = format_string $valence;
    if($objp) {
      $objp{$2.$3}{$1}{$objp}++;
    }
    if($objc) {
      $objc{$2.$3}{$1} = $objc;
    }
    if($loctype) {
      $loctype{$2.$3}{$1} = $loctype;
    }
    if($subjc) {
      $subjc{$2.$3}{$1} = 1;
    }
    $prefixes{$1}++;
    count_infinitive $1.$2.$3;
    return ('su',$1,$2,$3);
  }

  elsif ($decl =~ /^(.+)>(.+)<(.+)$/) {
    count_infinitive $1.$2.$3;
    return ('us',$1,$2,$3);
  }

  elsif ($decl =~ /^(.+)>(.+)$/) {
    count_infinitive $1.$2;
    return ('unstressed',$1,'',$2);
  }

  elsif ($decl =~ /^(.+)<(.+)$/) {
    my($pre, $stem) = ($1,$2);
    $avz{$stem}{$pre} = format_string $valence;
    $prefixes{$pre}++;
    count_infinitive $pre.$stem;
    if($objp) {
      if($objp =~ /^</) {
        while($objp =~ s/^<?([a-z]+),\d+[,>]//) {
          $objp{$stem}{$pre}{$1}++;
        }
      } else {
        $objp{$stem}{$pre}{$objp}++;
      }
    }
    if($objc) {
      $objc{$2}{$1} = $objc;
    }
    if($loctype) {
      $loctype{$2}{$1} = $loctype;
    }
    if($subjc) {
      $subjc{$2}{$1} = 1;
    }
    return ('stressed',$1,'',$2);
  }

  else {
    count_infinitive $decl;
    return ('none','','',$decl);
  }
}

# Parse a declaration line
# and fill @verbs with its data
our %decls;
sub read_declaration {
  my ($regularity, $decl, $valence, $extra) = @_;

  if($decls{$regularity.$decl}++) {
    warn "Duplicate verb: $decl!\n";
  }

  # force valence keys to be in alphabetical order, i.e. map P+D ==>
  # D+P. This halves the number of possible valence frames and
  # therefore allows us to get by with smaller hierarchies and
  # constraints.
  $valence = join('+',sort split(/\+/,$valence));

  # decode non-morphological info
  my $S = 'no';
  my %e = ( perfect => 'haben');
  if ($extra) {
    my @info = split(m| +|,$extra);
    for (@info) {

      if($_ eq 'Steigerung') {
        $S = 'yes';
        next;
      }

      /^([^:]+):(.+)$/ or die "Not a proper argument: `$_' (line $.)\n";
      my($key,$value) = ($1,$2);
      $e{$key} = $value;
    }
  }

  # decode morphological info
  my ($type,$prefix1,$prefix2,$base) =
    decode_declaration $decl, $valence, \%e;

  $e{valence} = $valence;
  for($valence) {
    if(!/^(-|r?[abcdefgikpnsx]+\??(\+r?[abcdefgikpnsx]+\??)?)$/) {
      die "Invalid valence frame `$_' (line $.)\n";
    }

    # make statistics
    $valences{$_}++;

    for(split(/\+/)) {
      s/[r\?]//g;
      $argtypes{$_}++;
    }
  }

  push @verbs,
   {
    base       => $base,
    class      => classify($base),
    extra      => \%e,
    inf        => $prefix1.$prefix2.$base,
    prefix1    => $prefix1,
    prefix2    => $prefix2,
    reg        => $regularity,
    type       => $type,
    Steigerung => $S
   };
}

# Print a lexical entry if it is wanted(). Also print regular variants
# of the form, i.e. strings that can be used interchangeably with the
# original form.
#
# There are two types of variants to verb forms:
#
# (1) Irregular variants. These must be declared with the `variant'
# keyword and apply to one form only. Example: `wurde' used to be
# `ward', so the input file should contain a declaration to that
# effect.
#
# (2) Regular variants. This are variants that are phonologically
# motivated and universal. For example, the imperative singular
# regularly drops the final `e' if there is one. This kind of variant
# need not be declared; it is applied automatically to all affected forms.
sub maybe_print {
  my $form = shift;
  my $entry = shift;
  my $alt = $form;

  # Generate form itself
  if (wanted($form)) {
    print format_string($form),  ":= $entry";
    $forms++;
  }

  # Generate irregular variants
  if(exists $variants{$form}) {
    for $alt(@{$variants{$form}}) {
      if (wanted $alt) {
        print format_string($alt), ":= $entry";
        $forms++;
      }
    }
  }

  # Generate regular variants

  # The stem e seems to be universally optional in certain contexts:
  # `verhandele' ==> `verhandle'
  if ($alt =~ s/([^aeiou])e([rl]e)$/$1$2/) {
    if (wanted $alt) {
      print format_string($alt), ":= $entry";
      $forms++;
    }
  }

  # Past tense forms in -dst have variants in -dest; in fact, those
  # are the more common ones.
  elsif ($entry =~ /tense:past/ &&
         $alt =~ s/([dt])st$/$1est/) {
    if (wanted $alt) {
      print format_string($alt), ":= $entry";
      $forms++;
    }
  }

  # Imperatives in -t have variants in -et
  elsif ($entry =~ /cat:VVIMP/ &&
         $alt =~ s/([^aeiou])t$/$1et/) {
    if (wanted $alt) {
      print format_string($alt), ":= $entry";
      $forms++;
    }
  }

  # Since 1999,  and ss are permissible variants in many situations,
  # so we simply allow both everywhere.
  $alt = $form;
  if ($alt =~ s/(e|en|end|ende|ende[mnrs]|t|te[nmrs]?|test|)$/ss$1/) {
    if (wanted $alt) {
      print format_string($alt), ":= $entry";
      $forms++;
    }
  } elsif ($alt =~ s/ss(e|en|t|te[nmrs]?|test|)$/$1/) {
    if (wanted $alt) {
      print format_string($alt), ":= $entry";
      $forms++;
    }
  }


  # Forms that end in `e' are often shortened:
  # `Ich habe ihn!' ==> `Ich hab' ihn!' / `Ich hab ihn!'
  if ($entry =~ /V[AV]IMP|V[AV]FIN.+present.+indicative/ && $form =~ /e$/) {
    (my $x = $form) =~ s/e$/\'/;
    if (wanted $x) {
      print format_string($x), ":= $entry";
      $forms++;
    }
    (my $y = $form) =~ s/e$//;
    if (wanted $y) {
      print format_string($y), ":= $entry";
      $forms++;
    }

    # Unfortunately the second and third rules can interact; forms that
    # end in `sse' can be modified to `e', and then further to `'. The
    # problem does not seem to warrant a full recursive generator, so we
    # just repeat the rules combined.
    if($form =~ /sse$/) {
      (my $z = $form) =~ s/sse$//;
      if (wanted $z) {
        print format_string($z), ":= $entry";
        $forms++;
      }
      ($z = $form) =~ s/sse$/\'/;
      if (wanted $z) {
        print format_string($z), ":= $entry";
        $forms++;
      }
    }
  }
}

# Map the string `sterben,stirbt,starb,strbe,gestorben'
# to a full hash of forms and note it in %paradigms.
sub make_paradigm {
  my $line = shift;
  my ($inf,$first,$second,$third,$past,$subj2,$ppp,$ppa);
  my $regular;
  my @paradigm = split(/,/,$line);

  # the paradigm for `sein' is so special, it has to be constructed explicitly.
  if ($line eq 'sein') {
    return $irregular_paradigms{sein} =
      {
       inf => 'sein',
       imp => { sg => { second => 'sei'}, pl => { second => 'seid' }},
       present =>
       { sg => { first => 'bin', second => 'bist', third  => 'ist' },
         pl => { first => 'sind', second => 'seid', third  => 'sind' }},
       past =>
       { sg => { first => 'war', second => 'warst', third  => 'war' },
         pl => { first => 'waren', second => 'wart', third  => 'waren' }},
       subj1 =>
       { sg => { first => 'sei', second => 'seiest', third  => 'sei' },
         pl => { first => 'seien', second => 'seid', third  => 'seien' }},
       subj2 =>
       { sg => { first => 'wre', second => 'wrst', third  => 'wre' },
         pl => { first => 'wren', second => 'wret', third  => 'wren' }},

       ppp => 'gewesen',
       ppa => 'seiend' };
  }

  # read infinitive
  $inf = $paradigm[0];

  my ($stem, $cons);

  if ($inf eq 'tun') {
    $stem = 'tu';
    $cons = ''
  } else {
    $inf =~ /^(.+?)e([lr]?)n$/
      or die "`$inf' doesn't look like an infinitive to me.\n";
    $stem = $1;

    # `subern', `stckeln'
    $cons = $2;
  }

  # note stems that affect inflection
  my $insert_e = 0;
  if ($stem =~ /(d|t|ch|ck|g|f|b)([mn])$/ && $cons eq '') {
    $insert_e = 1;
  }
  if ($stem =~ /[dt]$/ && $cons eq '') {
    $insert_e = 1;
  }

  my $elide_s = ($stem =~ /[s]$/);

  # Verbs with irregular 1st sg, 2nd sg, 3rd sg, past, subjunctive,
  # and ppp:
  # `halten,hltst,hlt,hielt,hielte,gehalten'
  if ($#paradigm == 6) {
    $regular = 0;
    ($inf,$first,$second,$third,$past,$subj2,$ppp) = @paradigm;
  }

  # Verbs with irregular 1st sg, 3rd sg, past, subjunctive, and ppp:
  # `wissen,wei,wei,wute,wte,gewut'
  elsif ($#paradigm == 5) {
    $regular = 0;
    ($inf,$first,$third,$past,$subj2,$ppp) = @paradigm;
  }

  # Verbs with irregular 3rd sg, past, subjunctive, and ppp:
  # `sterben,stirbt,starb,strbe,gestorben'
  elsif ($#paradigm == 4) {
    $regular = 0;
    ($inf,$third,$past,$subj2,$ppp) = @paradigm;
  }

  # Verbs with irregular past, subjunctive, and ppp:
  # `liegen,lag,lge,gelegen'
  elsif ($#paradigm == 3) {
    ($inf,$past,$subj2,$ppp) = @paradigm;
  }

  # regular verbs: `machen'
  elsif ($#paradigm == 0) {
    $regular = 1;
    $inf = $line;
  } else {
    die "Line: $line";
  }

  # compute the first person singular if it wasn't given
  if (!defined $first) {
    $first = $inf;
    $first =~ s/un$/ue/;     # tun -->tue
    $first =~ s/n$//;        # gehen --> gehe
  }

  # compute the third person singular if it wasn't given
  if (!defined $third) {
    $third = $inf;

    if ($insert_e) {
      $third =~ s/en$/et/;   # warten --> wartet, zeichnen --> zeichnet
    }

    # otherwise, substitute en -> t
    else {
      $third =~ s/un$/ut/;   # tun --> tut
      $third =~ s/en$/t/;    # gehen --> geht
    }
  }

  # compute the past stem if it wasn't given
  if (!defined $past) {
    $past = $inf;

    if ($cons ne '') {
      $past = "${stem}e${cons}te"; # sammelte, ruberte
    } elsif ($insert_e) {
      $past =~ s/en$/ete/;   # zeichnete, redete, wartete
    }

    # otherwise, substitute en -> te
    else {
      $past =~ s/en$/te/;    # machte
    }
  }

  # compute the subjunctive 2 if it wasn't given
  if (!defined $subj2) {
    $subj2 = $past;
  }

  # compute the ppp if it wasn't given
  if (!defined $ppp) {
    $ppp = $inf;
    # make exception for `visieren' but not `zieren'
    if ($inf =~ /[aeiou].*ieren$/) {
      $ppp =~ s/en$/t/;            # marschieren --> marschiert
    } else {
      $ppp =~ s/en$/t/;            # mhen --> mht
      $ppp =~ s/(e[lr])n$/$1t/;    # sammeln --> sammelt
      if ($insert_e) {
        $ppp =~ s/t$/et/;          # wartt --> wartet
      }
      $ppp =~ s/^/ge/;             # macht --> gemacht
    }
  }

  # compute the ppa
  $ppa = "${inf}d";

  # compute forms of the present tense
  my %present;
  if ($cons ne '') {
    my $pres = $stem . 'e' . $cons;
    %present =
      (
       sg =>
       { first => "${pres}e", second => "${pres}st", third => "${pres}t"},
       pl =>
       { first => "${pres}n", second => "${pres}t", third  => "${pres}n"} )
    } else {
      $present{sg}{first} = $first;

      if(!$second) {
        $second = $third;
        $second =~ s/(?<![s])t$/st/; # gibt --> gibst, lost --> lost
        $second =~ s/$/t/;          # wei --> weit
        $second =~ s/(?<!t)$/st/;     # darf --> darfst
        $second =~ s/wirdst$/wirst/;  # wirdst --> wirst
        $second =~ s/dst$/dest/;      # standst --> standest
        $second =~ s/itst$/ittst/;    # tritst --> trittst
      }
      $present{sg}{second} = $second;

      $present{sg}{third} = $third;

      $present{pl}{first} = $inf;
      $present{pl}{second} = $stem . ($insert_e ? 'e' : '') . 't';
      $present{pl}{third} = $inf;
    }

  # compute imperative

  # The singular imperative is sometimes built from the 2nd sg,
  # and sometimes from the 1st sg:
  #
  # ich gehe, du gehst ==> gehe!
  # ich trage, du trgst ==> trage!
  # ich spreche, du sprichst ==> sprich!
  #
  # If the 1st and 2nd do not differ in their vowels,
  # the imperative is identical to the 1sg.
  #
  # If the 2nd person has an umlaut where the 1st doesn't, then the
  # imperative follows the stem.
  #
  # If the 2nd has a different vowel that is not an umlaut, the
  # imperative has the new vowel.
  my $is;
  my $x1 = $present{sg}{first};
  my $x2 = $present{sg}{second};
  while (substr($x1,0,1) eq substr($x2,0,1)) {
    $x1 =~ s/^.//;
    $x2 =~ s/^.//;
  }
  # sichern ==> sichere!
  if ($cons) {
    $is = $present{sg}{first};
  }
  # trgst ==> trage!
  elsif (substr($x2,0,1) =~ /[]/ &&
         substr($x1,0,1) =~ /[aou]/) {
    $is = "${stem}e";
  }
  # brichst ==> brich!
  elsif (substr($x2,0,1) =~ /[aeiou]/ &&
         substr($x1,0,1) ne substr($x2,0,1)) {
    ($is = $present{sg}{third}) =~ s/t$//;
  }
  # gehst ==> gehe!
  else {
    $is = "${stem}e";
  }

  # The plural imperative is always identical to the plural indicative.
  my $ip = $present{pl}{second};
  my %imp = ( sg => {second => $is}, pl => {second => $ip} );


  # compute forms of the past tense
  my %past;
  my $past_ends_with_e = ($past =~ /e$/);
  my $past_ends_with_t = ($past =~ /([dt])$/);

  if ($regular || $cons ne '') {
    %past =
      (
       sg =>
       { first => $past, second => "${past}st", third => $past },
       pl =>
       { first => "${past}n",
         second => $past . ($past_ends_with_e ? '' : 'e') . 't',
         third => "${past}n" }
      );
  } else {
    %past =
      (
       sg =>
       { first => $past, second => "${past}st", third => $past },
       pl =>
       { first => $past . ($past_ends_with_e ? '' : 'e') . 'n',
         second => $past . ($past_ends_with_t ? 'e' : '') . 't',
         third  => $past . ($past_ends_with_e ? '' : 'e') . 'n'}
      );
  }

  # compute forms of the subjunctive
  my %subj1;
  if ($cons ne '') {
    my $subj1 = "${stem}e$cons";
    %subj1 =
      (
       sg =>
       { first => "${subj1}e", second => "${subj1}st", third => "${subj1}e" },
       pl =>
       { first => "${subj1}n", second => "${subj1}t", third => "${subj1}n"}
      )
    } else {
      %subj1 =
        (
         sg =>
         { first => "${stem}e", second => "${stem}est", third => "${stem}e"},
         pl =>
         { first => "${stem}en", second => "${stem}et", third => "${stem}en"}
        );
    }

  my %subj2 =
    (
     sg =>
     { first => $subj2, second => "${subj2}st", third => $subj2},
     pl =>
     { first => "${subj2}n", second => "${subj2}t", third => "${subj2}n" }
    );

  my $result =
    {
     inf => $inf,
     imp => \%imp,
     present => \%present,
     past => \%past,
     subj1 => \%subj1,
     subj2 => \%subj2,
     ppp => $ppp,
     ppa => $ppa
    };

  if ($#paradigm == 0) {
    return $regular_paradigms{$inf} = $result;
  } else {
    return $irregular_paradigms{$inf} = $result;
  }
}

# Retrieve a paradigm, constructing it if necessary.
sub get_paradigm {
  my $inf = shift;
  if (exists $irregular_paradigms{$inf}) {
    return $irregular_paradigms{$inf};
  } elsif (exists $regular_paradigms{$inf}) {
    return $regular_paradigms{$inf};
  }
  return make_paradigm($inf);
}

# get the regular paradigm for an infinitive,
# even if an irregular one exists.
sub get_regular_paradigm {
  my $inf = shift;
  if(exists $regular_paradigms{$inf}) {
    return $regular_paradigms{$inf};
  }
  return make_paradigm($inf);
}

# Dump a paradigm to screen
sub print_paradigm {
  my $p = shift;
  print "$p->{inf}\n\n";
  for (qw(present past subj1 subj2)) {
    print "$_: ";
    print "ich $p->{$_}{sg}{first}, ";
    print "du $p->{$_}{sg}{second}, ";
    print "er $p->{$_}{sg}third}, ";
    print "wir $p->{$_}{pl}{first}, ";
    print "ihr $p->{$_}{pl}{second}, ";
    print "sie $p->{$_}{pl}{third}\n";
  }
  print "PPP: $p->{ppp}\n";
  print "PPA: $p->{ppa}\n\n";
}

# Return list of paradigms for a verb,
# depending on its regularity qualifier
sub find_paradigms {
  my $verb = shift;
  my $regularity = shift;
  my @result;

  if($regularity eq 'regular') {
    push @result, get_regular_paradigm($verb);
  }

  elsif($regularity eq 'semiregular') {
    push @result, get_paradigm($verb);
    push @result, get_regular_paradigm($verb);
  }

  else {
    push @result, get_paradigm($verb);
  }

  return @result;
}

# There are instances where a certain verb takes an object of a
# particular type, but only in coordination with an AVZ. In this case,
# whether or not the object is legal depends on the presence and type
# of the AVZ, and we must enforce the condition via a binary
# constraint.
#
# But we want to write another, more efficient constraint that forbids
# those object types that are not possible even with an AVZ, without
# using lookup() or has(). For that purpose, we collect the
# additionally possible object types and project them onto the base
# verb as the feature `extravalence'.
#
# An example should make things clearer. Suppose the verb `kosten'
# takes an OBJA and an OBJA2. Its valence of 'A+B' reflects this. The
# separable verb `aus<kosten' takes only a normal OBJA. This is easy
# to model; if we find the edge aus--AVZ-->koste, then we know that we
# have to check the valence of `auskosten', not of `kosten', and
# therefore OBJA is still allowed, but not OBJA2.
#
# But what if the opposite is true? Consider `schlagen', which takes a
# single OBJA. The complex verb `vor<schlagen' takes an OBJA *or* an
# OBJC, so the edge auszusteigen--OBJC-->schlage cannot be ruled out
# -- there might be a `vor' following. But there are other, really
# impossible edges, such as sich--OBJG-->verkaufen. This is total
# nonsense; `verkaufen' doesn't take a genitive, and although it has
# frame-shifting variants, none of them takes a genitive either. In
# this case, we want to write a unary constraint that forbids the edge
# immediately. It is this constraint that will check the value of
# X@extravalence; cf. the constraint `falsches <OBJX>'.
#
# A very similar line of reasoning holds for extraobjp and extrasubjc.
sub compute_extra_valence {

  my $v = shift;

  my $resv = '';
  my $resp = '';
  my $ress = 0;

  my $result = '';

  # compute valences only allowed with AVZ
  my $oldv = $$v{extra}{valence};
  if (exists $avz{$$v{inf}}) {
    for my $newv (values %{$avz{$$v{inf}}}) {
      for my $l (keys %labels) {
        if ($newv =~ /$l/ &&
            $oldv !~ /$l/ &&
            $resv !~ /$l/) {
          $resv .= $l;
        }
      }
    }
  }

  # compute OBJP types only allowed with AVZ
  my $oldp = $$v{extra}{objp};
  my %oldps;
  my %newps;
  if($oldp) {
    while ($oldp =~ s/([a-z]+)//) {
      $oldps{$1}++;
    }
  }
  for my $newp (values %{$objp{$$v{inf}}}) {
    for my $n(keys %$newp) {
      while ($n =~ s/([a-z]+)//) {
        if (!exists($oldps{$1})) {
          $newps{$1}++;
        }
      }
    }
  }

  if(%newps) {
    $resp = '';
    my @l = keys %newps;
    if(@l == 1) {
      $resp .= "$l[0]";
    } else {
      my $first = 0;
      $resp .= '<';
      for(@l) {
        $resp .= ',' if $first++;
        $resp .= "$_,1";
      }
      $resp .= '>';
    }
  }

  # detect AVZ variants that take unmarked subject clauses
  if(exists($subjc{$$v{inf}})) {
    $ress = 1;
  }

  if($resv) {
    $valences{$resv}++;
    $result .= ",extravalence:$resv";
   }

  if($resp) {
    $result .= ",extraobjp:$resp";
  }

  if($ress) {
    $result .= ",extrasubjc:yes";
  }

  return $result;
}

# Maps a hash of features to a string suitable for appearing within a
# lexicon entry.
#
# Example: { valence => 'A+D', perfect => 'sein' } is mapped to
# the string "valence:'A+D',perfect:sein".
sub print_features {
  my $f = shift;
  my $result;
  while (my($k,$v) = each %$f) {

    # do not quote AV lists
    if($v !~ /^<.+>$/) {
      $v = format_string $v;
    }
    $result .= ",$k:$v";
  }
  return $result;
}

# Generate string of morphology information a lexicon entry.
#
# Example: { p =>'second', n => 'pl' } is mapped to the string
# "person:third,number:pl".
sub print_morph {
  my $f = shift;
  my $result;
  $result .= "person:$f->{p}," if exists $f->{p};
  $result .= "number:$f->{n}," if exists $f->{n};
  if($f->{tm} eq 'present') {
    $result .= 'tense:present,mood:indicative';
  }
  elsif($f->{tm} eq 'past') {
    $result .= 'tense:past,mood:indicative';
  }
  elsif($f->{tm} eq 'subj1') {
    $result .= 'tense:present,mood:subjunctive1';
  }
  elsif($f->{tm} eq 'subj2') {
    $result .= 'tense:present,mood:subjunctive2';
  }
  elsif($f->{tm} eq 'imp') {
    $result .= 'mood:indicative';
  }
  else {
    die "Invalid tense/mood marker: `$f->{tm}'!\n";
  }
  return $result;
}

# Generate a lexical entry for the VINF form
sub print_VINF {
  my $v = shift;

  my $form = $$v{prefix1}.$$v{prefix2}.$$v{base};

  my $entry = "[ infinitive:" .
    format_string($form) . ", cat:V$$v{class}INF";
  $entry .= print_features $$v{extra};
  $entry .= " ];\n\n";

  maybe_print($form,$entry);
}

# Generate a lexical entry for the VIZU form
sub print_VIZU {
  my $v = shift;

  # Only some verb classes have an IZU variant at all.
  my $form;
  if($$v{type} eq 'stressed') {
    $form = $$v{prefix1} . 'zu' . $$v{base};
  }
  elsif($$v{type} eq 'su') {
    $form = $$v{prefix1} . 'zu' . $$v{prefix2} . $$v{base};
  }
  else {
    return;
  }

  my $inf = format_string $$v{inf};
  my $entry = "[ infinitive:$inf, cat:V$$v{class}IZU ";
  $entry .= print_features $$v{extra};
  $entry .= "];\n\n";

  maybe_print($form,$entry);
}


# Generate lexical entry for a VVFIN form.
#
# The hash %$v contains all information about the particular verb, the
# hash %$morph about the actual form. $skip_SVO is set if we don't
# need separated SVO forms (see %infinitives).
sub print_VVFIN {
  my $v = shift;
  my $morph = shift;
  my $skip_SVO = shift;
  my $form = shift;
  my $extra_valence = shift;

  # find form
  my @forms = ( $form );

  my $inf = format_string $$v{inf};

  for my $form (@forms) {

    # catch imperatives
    my $cat = "V$$v{class}FIN";
    if ($morph->{tm} eq 'imp') {

      # modal imperatives do not exist
      return if($$v{class} eq 'M');

      # imperatives are a separate POS class
      $cat = "V$$v{class}IMP";
    }

    my $entry;

    # Forms without prefixes are simplest.
    if ($$v{type} eq 'none') {
      my $avz_status = 'forbidden';
      if(exists($avz{$$v{base}})) {
        $avz_status = 'allowed';
       }
      $entry = "[ infinitive: $inf, cat:$cat, avz:$avz_status,";
      $entry .= print_morph $morph;
      $entry .= print_features $$v{extra};
      $entry .= $extra_valence;
      $entry .= "];\n\n";
      maybe_print($form,$entry);
    }

    # Stressed prefixes behave differently in main clauses and
    # subclauses, therefore we actually need two lexical entries. But
    # we only print the separated form if it is not available on its
    # own.
    elsif ($$v{type} eq 'stressed') {
      if ($morph->{tm} ne 'imp') {
        my $f = "$$v{prefix1}$form";
        $entry = "[ infinitive: $inf, cat:$cat,pos:SOV,avz:forbidden,";
        $entry .= print_morph $morph;
        $entry .= print_features $$v{extra};
        $entry .= "];\n\n";
        maybe_print($f,$entry);
      }

      if(!$skip_SVO) {
        $entry = "[ infinitive: $$v{base}, cat:$cat,pos:SVO,avz:required,";
        $entry .= print_morph $morph;
        $entry .= print_features $$v{extra};
        $entry .= "];\n\n";
        maybe_print($form, $entry);
      }
    }

    # Unstressed prefixes simply attach to the finite form no matter what.
    elsif ($$v{type} eq 'unstressed') {
      my $f = "$$v{prefix1}$form";

      my $avz_status = 'forbidden';
      if(exists($avz{$$v{prefix1}.$$v{base}})) {
        $avz_status = 'allowed';
      }

      $entry = "[ infinitive: $inf, cat:$cat,avz:$avz_status,";
      $entry .= print_morph $morph;
      $entry .= print_features $$v{extra};
      $entry .= $extra_valence;
      $entry .= "];\n\n";
      maybe_print($f,$entry);
    }

    # Multiple prefixes where the first is stressed
    # behave like stressed prefixes.
    elsif ($$v{type} eq 'su') {
      if ($morph->{tm} ne 'imp') {
        my $f = "$$v{prefix1}$$v{prefix2}$form";
        $entry = "[ infinitive: $inf, cat:$cat,pos:SOV,avz:forbidden,";
        $entry .= print_morph $morph;
        $entry .= print_features $$v{extra};
        $entry .= "];\n\n";
        maybe_print($f,$entry);
      }

      if (!$skip_SVO) {
        my $f = "$$v{prefix2}$form";
        $entry = "[ infinitive:$$v{prefix2}$$v{base},";
        $entry .= "cat:$cat,pos:SVO,avz:required,";
        $entry .= print_morph $morph;
        $entry .= print_features $$v{extra};
        $entry .= "];\n\n";
        maybe_print($f,$entry);
      }
    }

    # Multiple prefixes where the second is stressed
    # behave like unstressed prefixes.
    elsif ($$v{type} eq 'us') {
      my $f = "$$v{prefix1}$$v{prefix2}$form";
      $entry = "[ infinitive: $inf, cat:$cat,avz:forbidden,";
      $entry .= print_morph $morph;
      $entry .= print_features $$v{extra};
      $entry .= "];\n\n";
      maybe_print($f,$entry);
    }
  }
}

# Generate all finite forms for a verb.
sub print_finite_forms {

  my $v = shift;

  my $skip_SVO = 0;

  if($$v{type} eq 'stressed') {
    if(exists $infinitives{$$v{base}}) {
      $skip_SVO = 1;
    }
  }
  elsif($$v{type} eq 'su') {
    if(exists $infinitives{$$v{prefix2}.$$v{base}}) {
      $skip_SVO = 1;
    }
  }

  my @p = find_paradigms($$v{base}, $$v{reg});

  my $extravalence = compute_extra_valence $v;
 
  for(my $i = 0; $i <= $#p; $i++) {

    # imperatives
    for my $n qw(sg pl) {
      my $form = $p[$i]{imp}{$n}{second};

      # Suppress alternative form
      # if it is identical to the standard form
      next if $i == 1 && $form eq $p[0]{imp}{$n}{second};

      print_VVFIN($v, { tm=>'imp', n=>$n, p=>'second' }, $skip_SVO, $form, $extravalence);
    }
  }

  # generate finite forms
  my %forms;
  for(my $i = 0; $i <= $#p; $i++) {
    for my $tm qw(present past subj1 subj2) {
      for my $n qw(sg pl) {
        for my $p qw(first second third) {
          my $form = $p[$i]{$tm}{$n}{$p};

          # Suppress alternative form
          # if it is identical to the standard form
          next if $i == 1 && $form eq $p[0]{$tm}{$n}{$p};

          push @{$forms{$form}}, { tm => $tm, n => $n, p => $p };
        }
      }
    }
  }

  # print them
  for my $f (sort keys %forms) {
    my @a = @{$forms{$f}};
    for my $i(@a) {
      print_VVFIN($v, $i, $skip_SVO, $f, $extravalence);
    }
  }

  if($$v{type} eq 'stressed') {
    $infinitives{$$v{base}}++
  }
  elsif($$v{type} eq 'su') {
    $infinitives{$$v{prefix2}.$$v{base}}++;
  }
}

# Generate semi-instantiated morphosyntactical features based on the
# ending of an adjective.
sub morphology {

  my $form = shift;
  if($form =~ /er$/) {
    return ", case:nom_gen_dat,number:bot,gender:bot,flexion:strong_mixed,suffix:er";
  } elsif ($form =~ /e$/) {
    return ", case:nom_acc,number:bot,gender:bot,flexion:bot,suffix:e";
  } elsif ($form =~ /em$/) {
    return ", case:dat,number:sg,gender:not_fem,flexion:strong,suffix:em";
  } elsif ($form =~ /es$/) {
    return ", case:nom_acc,number:sg,gender:neut,flexion:strong_mixed,suffix:es";
  } elsif ($form =~ /en$/) {
    return ", case:bot,number:bot,gender:bot,flexion:bot,suffix:en";
  }
}

# Generate a lexical entry for the passive verbal adjective
sub print_passive_ADJA {
  my $form = shift;
  my $v = shift;
  my $degree = shift;

  my $entry = "[cat: ADJA, infinitive:" . format_string($$v{inf})
    . ", partizipial2:yes, degree:$degree";
  $entry .= print_features $$v{extra};

  $entry .= morphology($form);
  $entry .= "];\n";

  maybe_print($form,$entry);
}

# Generate a lexical entry for the PP form.
sub print_PP {
  my $v = shift;

  # find one or two possible forms
  my @p = find_paradigms($$v{base}, $$v{reg});
  my @forms = ($p[0]{ppp});
  if($#p == 1) {
    my $alt = $p[1]{ppp};
    push @forms, $alt if $alt ne $p[0]{ppp};
  }

  for my $form(@forms) {

    if ($$v{type} eq 'stressed') {
      $form = $$v{prefix1} . $form;
    } elsif ($$v{type} eq 'unstressed') {
      $form =~ s/^ge//;
      $form = $$v{prefix1} . $form;
    } elsif ($$v{type} eq 'su') {
      $form =~ s/^ge//;
      $form = $$v{prefix1} . $$v{prefix2} . $form;
    } elsif ($$v{type} eq 'us') {
      $form =~ s/^ge//;
      $form = $$v{prefix1} . $$v{prefix2} . $form;
    }

    my $entry = "[ infinitive:" .
      format_string("$$v{prefix1}$$v{prefix2}$p[0]{inf}") .
        ",cat:V$$v{class}PP";
    $entry .= print_features $$v{extra};
    $entry .= "];\n\n";
    maybe_print($form,$entry);

    # adverbial usage
    $entry = "[ infinitive:" .
      format_string("$$v{prefix1}$$v{prefix2}$p[0]{inf}") .
        ",cat:ADJD";
    $entry .= ", partizipial2:yes, degree:positive";
    $entry .= print_features $$v{extra};
    $entry .= "];\n\n";
     maybe_print($form,$entry);

    if($$v{Steigerung} eq 'yes') {
      $entry = "[ infinitive:" .
        format_string("$$v{prefix1}$$v{prefix2}$p[0]{inf}") .
          ",cat:ADJD";
      $entry .= ", partizipial2:yes, degree:comparative";
      $entry .= print_features $$v{extra};
      $entry .= "];\n\n";
      maybe_print("${form}er",$entry);
      $entry = "[ infinitive:" .
        format_string("$$v{prefix1}$$v{prefix2}$p[0]{inf}") .
          ",cat:ADJD";
      $entry .= ", partizipial2:yes, degree:superlative";
      $entry .= print_features $$v{extra};
      $entry .= "];\n\n";
      maybe_print("${form}esten",$entry);
    }

    # `verkauftes' etc.
    print_passive_ADJA("${form}e", $v, 'positive');
    print_passive_ADJA("${form}er", $v, 'positive');
    print_passive_ADJA("${form}en", $v, 'positive');
    print_passive_ADJA("${form}em", $v, 'positive');
    print_passive_ADJA("${form}es", $v, 'positive');

    if($$v{Steigerung} eq 'yes') {
      print_passive_ADJA("${form}ere", $v, 'comparative');
      print_passive_ADJA("${form}erer", $v, 'comparative');
      print_passive_ADJA("${form}eren", $v, 'comparative');
      print_passive_ADJA("${form}erem", $v, 'comparative');
      print_passive_ADJA("${form}eres", $v, 'comparative');

      print_passive_ADJA("${form}este", $v, 'superlative');
      print_passive_ADJA("${form}ester", $v, 'superlative');
      print_passive_ADJA("${form}esten", $v, 'superlative');
      print_passive_ADJA("${form}estem", $v, 'superlative');
      print_passive_ADJA("${form}estes", $v, 'superlative');
    }
  }
}

# Generate a lexical entry for the active verbal adjective.
sub print_active_ADJA {
  my $form = shift;
  my $cat = shift;
  my $features = shift;
  my $degree = shift;

  my $entry = "[cat: ADJ$cat, partizipial1:yes, degree:$degree";
  $entry .= print_features $features;
  if($cat eq 'A') {
    $entry .= morphology $form;
  }
  $entry .= "];\n";

  maybe_print($form,$entry);
}

# Generate verbal adjectives
sub print_adjectives {
  my $v = shift;

  my $p = get_paradigm $$v{base};

  # `verkaufendes' etc.
  my $ppa = $$v{prefix1} . $$v{prefix2} . $p->{ppa};

  print_active_ADJA("$ppa", 'D', $$v{extra}, 'positive');

  for(qw(e er en em es)) {
    print_active_ADJA("${ppa}$_", 'A', $$v{extra}, 'positive');

    if(($$v{type} eq 'stressed' || $$v{type} eq 'su') &&
      $$v{extra}{valence} =~ /a/) {
      my $f = $$v{prefix1} . 'zu' . $$v{prefix2} . $$p{ppa};
      print_active_ADJA("${f}$_", 'A', $$v{extra}, 'positive');
    }

  }

  # Only some verbs form comparative and superlative forms of the
  # verbal adjective.
  if($$v{Steigerung} eq 'yes') {
    print_active_ADJA("${ppa}er", 'D', $$v{extra}, 'comparative');
    print_active_ADJA("${ppa}sten", 'D', $$v{extra}, 'superlative');

    for(qw(e er en em es)) {
      print_active_ADJA("${ppa}er$_", 'A', $$v{extra}, 'comparative');
      print_active_ADJA("${ppa}st$_", 'A', $$v{extra}, 'superlative');
    }
  }
}

sub make_verb {

  my $v = shift;

  # detect verbs that are ambiguous w.r.t. stress
  if(exists $ambiguous_stress{$$v{inf}}) {
    $$v{extra}{stress} = $$v{type};
  }

  # infinitives
  print_VINF($v);
  print_VIZU($v);

  # finite forms
  print_finite_forms($v);

  # participles
  print_PP($v);

  # verbal adjectives
  print_adjectives($v);

}

# generate the hierarchies necessary for argument matching
sub write_valence_hierarchy {

  sub find_subs {
    my($name, $re) = @_;
    my $count = 0;
    print OUT "$name -> ";
    for(keys %valences) {
      ($count++, print OUT "'$_' ") if $_ =~ $re;
    }
    print OUT 'bot' unless $count;
    print OUT ",\n";
  }

  my @types = keys %labels;

  # These valence frames happen to occur in some nouns and adjectives
  # but in no verbs. Since this program is responsible for creating
  # the entire Valence subsumption table, we must list them by hand.
  # This would not be necessary if the various make-stuff.pl's were
  # unified.
  $valences{'cgp?'}++;
  $valences{'ci?'}++;
  $valences{'cip?'}++;
  $valences{'cis?'}++;
  $valences{'cisg?'}++;
  $valences{'cs?'}++;
  $valences{'csg'}++;
  $valences{'d?+ga?'}++;
  $valences{'d?+ci?'}++;
  $valences{'rd?+ci?'}++;
  $valences{'rd?'}++;
  $valences{'g?'}++;
  $valences{'ga?'}++;
  $valences{'gc?'}++;
  $valences{'gi?'}++;
  $valences{'igp?'}++;
  $valences{cg}++;

  open(OUT, ">AVZ.cdg") or die "Can't write `AVZ.cdg': $!\n";

  print OUT "\n\nValenzen ->\n";

  for my $t (@types) {
    find_subs "${t}_erlaubt", qr/$t/;
  }

  # FIXME: automate this as well
  find_subs 'D_und_P_erlaubt', qr/(d.*\+.*p)|(p.*\+.*d)/;
  find_subs 'A_und_P_erlaubt', qr/(a.*\+.*p)|(p.*\+.*a)/;
  find_subs 'A_und_C_erlaubt', qr/(a.*\+.*c)|(c.*\+.*a)/;
  find_subs 'A_und_I_erlaubt', qr/(a.*\+.*i)|(i.*\+.*a)/;
  find_subs 'A_und_S_erlaubt', qr/(a.*\+.*s)|(s.*\+.*a)/;
  find_subs 'C_und_P_erlaubt', qr/(p.*\+.*[cis])|([cis].*\+.*p)/;

  find_subs 'A_notwendig', qr/a[^?+]*(\+|$)/;
  find_subs 'C_notwendig', qr/c[^?+]*(\+|$)/;
  find_subs 'S_notwendig', qr/s[^?+]*(\+|$)/;

  find_subs 'A_nur_reflexiv', qr/ra/;
  find_subs 'D_nur_reflexiv', qr/rd/;

  find_subs 'transitiv', qr/^[a-z]+(\+|$)/;
  find_subs 'bitransitiv', qr/\+[a-z]+$/;

  print OUT "'-' -> bot, '' -> bot\n";
  print OUT ";\n\n";

  # generate hierarchies that translate from valence strings to
  # lists of permissible labels
  print OUT "Valenz_1 ->\n";
  for(keys %valences) {
    next if $_ eq '-';
    print OUT "'$_' ->";
    # ignore second frame
    s/\+.+$//;
    while(s/([a-z])//) {
      my $label = $labels{$1};
      if($label) {
        print OUT " $label";
      }
    }
    print OUT ",\n";
  }
  print OUT "'-' -> bot, '' -> bot\n";
  print OUT ";\n\n";

  print OUT "Valenz_2 ->\n";
  for(keys %valences) {
    next if $_ eq '-';
    next unless /\+/;
    print OUT "'$_' ->";
    # ignore first frame
    s/^.+\+//;
    while(s/([a-z])//) {
      my $label = $labels{$1};
      if($label) {
        print OUT " $label";
      }
    }
    print OUT ",\n";
  }
  print OUT "'-' -> bot, '' -> bot\n";
  print OUT ";\n\n";

  close(OUT);

}

# generate constraints about verb valences
sub write_constraints {

  open(OUT, ">>AVZ.cdg")
    or die "Can't write `AVZ.cdg': $!\n";

  # Init constraints

  # Each of the verb-complement labels is only allowed with a verb
  # that explicitly subcategorises for this complement. But since
  # verbs can change their frame when they take an AVZ, we have to
  # distinguish between verbs with and without an AVZ.

  while (my($k,$v) = each %labels) {

    # KOM is not strictly a complement and can appear everywhere
    next if $v eq 'KOM';

    # And so can OBJD
    next if $v eq 'OBJD';

    # PN is mostly allowed by adpositions, so we don't repeat that
    # constraint here; instead we add another for verbs manually in
    # Grammatik.cdg.
    if($v ne 'PN') {

      my $maybe_allow_PDS = $v =~ /OBJ[CI]/ ? 'X^cat = PDS |' : '';

      # When no AVZ can be present, allowed objects are distinguished
      # by checking X^valence.
      print OUT <<EOT;
{X!SYN} : '${v} nicht erlaubt' : init : 0.0 :
  X.label = $v
  ->
  $maybe_allow_PDS
  exists(X^valence) & subsumes(Valenzen, ${k}_erlaubt, X^valence) |
  exists(X^avz) & X^avz != forbidden;

EOT
}
    # When an AVZ is present, we have to look up the valence
    # of prefix and verb jointly.
    print OUT <<EOT;
{X!SYN/\\Y!SYN} : '${v} und AVZ nicht erlaubt' : init : 0.0 :
  Y.label = AVZ &
  X.label = $v
  ->
  exists(X^valence) &
  subsumes(Valenzen,
           ${k}_erlaubt,
           lookup(AVZ, Y\@word, X^infinitive));

EOT

    # When the object is only allowed with an AVZ,
    # and the AVZ is not present, the object is not allowed.
    print OUT <<EOT;
{X!SYN} : '${v} nur mit AVZ erlaubt' : init : 0.0 :
  X.label = $v &
  isa(X^,Verb) &
  exists(X^extravalence) &
  subsumes(Valenzen, ${k}_erlaubt, X^extravalence) &
  ~has(X^id,AVZ)
  ->
  exists(X^valence) &
  subsumes(Valenzen, ${k}_erlaubt, X^valence);

EOT

    # When the object is not allowed with or without an AVZ,
    # it is definitely not allowed. In this case we can forbid
    # it without using has(), so we do it, because a unary constraint
    # is much more effective than a unary context-sensitive constraint.
    print OUT <<EOT;
{X!SYN} : 'falsches ${v}' : init : 0.0 :
  X.label = $v &
  exists(X^avz) & X^avz = allowed &
  ~subsumes(Valenzen, ${k}_erlaubt, X^valence)
  ->
  exists(X^extravalence) &
  subsumes(Valenzen, ${k}_erlaubt, X^extravalence);

EOT
  }

  # FIXME: co-occurrence constraints

  close(OUT);
}

# generate [cat:PTKVZ] forms and AVZ table
sub write_prefixes {

  open(OUT, ">>AVZ.cdg")
  or die "Can't write `AVZ.cdg': $!\n";

  # make any prefixes necessary for these verbs
  for (sort keys %prefixes) {

    printf OUT "%-15s := [cat:TRUNC,cat2:Verb];\n", format_string "$_-";

    # do not generate items of these prefixes, which are really either
    # adverbs or postpositions.
    next if /^((hin|her)(auf|ab|unter)|entlang)$/;

    printf OUT "%-15s := [cat:PTKVZ];\n", format_string $_;
  }

  # write AVZ table
  print OUT "\n\nAVZ =>\n";
  my $first = 0;
  for my $v (sort keys %avz) {
    for my $p(sort keys %{$avz{$v}}) {
      my $fp = format_string $p;
      print OUT ",\n" if $first++;
      my $val = $avz{$v}{$p};
      print OUT "  $fp $v => $val";

      for my $objp(sort keys %{$objp{$v}{$p}}) {
        print OUT ",\n    $fp $v $objp => ok";
      }
      if(exists $objc{$v}{$p}) {
        print OUT ",\n    $fp $v objc => $objc{$v}{$p}";
      }
      if(exists $loctype{$v}{$p}) {
        print OUT ",\n    $fp $v loctype => $loctype{$v}{$p}";
      }
      if(exists $subjc{$v}{$p}) {
        print OUT ",\n    $fp $v subjc => yes";
      }
    }
  }

  print OUT ";\n\n\n";
  close(OUT);
}

# MAIN

while (<>) {
  chomp;

  # trim whitespace
  s/^\s+//;
  s/\s+$//;

  # find comments
  next if /^\ *\#/;
  next if /^\s*$/;

  # find macros
  if (/^macro +([A-Z]+) +([^ ]+)$/) {
    make_macro $1, $2;
    next;
  }

  # expand macros
  $_ = expand_macros $_;

  # find paradigms
  if (/^paradigm +(.+)$/) {
    make_paradigm $1;
  }

  # find verb declaration
  elsif (/^(semiregular|regular|)\s*verb\s+(\S+)\s+(\S+)(\s+(.+))?$/) {
    read_declaration($1, $2, $3, $5);
  }

  # find variant declaration
  elsif (/^variant\s+(\S+)\s+(\S+)\s*$/) {
    push @{$variants{$1}}, $2;
  }

  # reject other lines
  else {
    die "Invalid line $.: `$_'\n";
  }
}

for (@verbs) {
  make_verb $_;
}
write_valence_hierarchy;
write_prefixes;
write_constraints;

print STDERR "Generated $forms verb forms.\n";

