# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## -----------------------------------------------------------------------------
## CommandHistory - implements generic undo-functionality 
## (used for editing Parse-Trees, see class ParseTree and VisParses)
##
##
## \author Dietmar Dreyer
##
## $Id commandhistory.tcl $
##-------------------------------------------------------------------------------

class CommandHistory {

    # private variables

    ## current position in command lists
    private variable _cursor -1 ; # \type TclNumber

    ## list of commands used for redoing actions
    private variable _commandList "" ;# \type TclList

    ## list of commands used for undoing actions
    private variable _inverseCommandList "" ; # \type TclList

    ## list of object-IDs recognizing update-method
    private variable _listenerList "" ; # \type TclList

    constructor {} {}

    # public methods
    public method add {command commandInverse}; ## \type TclList, TclList
    public method undo {} 
    public method redo {}
    public method canUndo {}
    public method canRedo {}
    public method addListener {listener}; ## \type TclString

    # protected methods
    protected method update {}

}

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body CommandHistory::constructor {} {

}


##-------------------------------------------------------------------------------
## Adds command for redoing and undoing current action to the end of command-list
## @param command command for redoing
## @param inverseCommand command for undoing
##-------------------------------------------------------------------------------
body CommandHistory::add {command inverseCommand} {
    set length [llength $_commandList];

    if { $length > [expr $_cursor]} {
	set _commandList [lrange $_commandList 0 $_cursor]
	set _inverseCommandList [lrange $_inverseCommandList 0 $_cursor]
    }

    lappend _commandList $command
    lappend _inverseCommandList $inverseCommand
    set _cursor [expr [llength $_commandList] -1]

    update
}


##-------------------------------------------------------------------------------
## Internal method used to update registered listeners whenever a change to the
## list of commands occurred
##-------------------------------------------------------------------------------
body CommandHistory::update {} {
    foreach listener $_listenerList {
	$listener update
    }

}


##-------------------------------------------------------------------------------
## Registers listener object which will be called whenever a change to the 
## command list happens 
## @param listener Object-ID implementing method update
##-------------------------------------------------------------------------------
body CommandHistory::addListener { listener } {
    if { [lsearch $_listenerList $listener] == -1} {
	lappend _listenerList $listener
    }
}



##-------------------------------------------------------------------------------
## Calls current undo command and modifies internal cursor to point to previous
## command in the list (move to beginning)
##-------------------------------------------------------------------------------
body CommandHistory::undo {} {

    if { $_cursor >= 0 } {
	eval [lindex $_inverseCommandList $_cursor]
	set _cursor [expr $_cursor-1]
    }

    update
}


##-------------------------------------------------------------------------------
## Calls current redo command and let cursor point to next command in the list
## (move to end)
##-------------------------------------------------------------------------------
body CommandHistory::redo {} {
    
    if { $_cursor < [expr [llength $_commandList] -1]} {
	set _cursor [expr $_cursor + 1]
	eval [lindex $_commandList $_cursor]  
    }
    update
}


##-------------------------------------------------------------------------------
## Predicate to determine if calling undo will take effect
## @return 1 => undo will take effect, 0 => not possible 
##-------------------------------------------------------------------------------
body CommandHistory::canUndo {} {

    if { $_cursor >= 0} {
	return 1
    } else {
	return 0
    }

}


##-------------------------------------------------------------------------------
## Similar to canUndo
## @return 1 => redo possible, 0 => no redo possible
##-------------------------------------------------------------------------------
body CommandHistory::canRedo {} {

    if { $_cursor < [expr [llength $_commandList] -1]} {
	return 1
    } else {
	return 0
    }
}
