# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## DataBrowser - Abstract base class for tabbed panes containing 
## a top frame of buttons and a center table
##
##
## Rules for subclassing:
##
## - Override init_data in order to perform any additional initializations
##
## - Call init_data whenever the frontend has to be set back to an initial
##   state, e.g. after loading a new file 
##
## - Always call init_data at the end of derived constructor just before
##   itk_initialize 
##
## - Add and pack additional components in derived constructor 
##   use component childsite to insert buttons into top panel
##
## - All remaining methods possess default implementation, which may be
##   overridden 
##
## - Use variable _selection to store table row selection ids
##
## - Set variable _idColumnIndex to index number of column, that will contain 
##   selectable IDs (set at beginning of derived constructor!)
##
## - Override getCData to get a pointer to the underlying C-Structure
## - Override refreshrow to get data corresponding to the pointer retrieved by getCData
##   and writing it into the table
##
## \author Dietmar Dreyer (see also AUTHORS and THANKS for more)
##
## $Id: databrowser.tcl,v 1.13 2004/10/11 13:50:06 micha Exp $
## ----------------------------------------------------------------------------

class DataBrowser {
   inherit ::itk::Widget

    # public methods
    public method getSelection {}
    public method setSelection {args}; ## \type TclList
    public method setIndexedSelection {args}; ## \type TclList

    ## abstract methods for retrieving the relevant C or Tcl data values for a given ID 
    public method getCData {id}; ## \type TclString

    public method refreshid {id}; ## \type TclString

    # protected methods
    protected method init_data {}; ## call at the end of derived constructor
    protected method _rowtag {row}; ## \type TclNumber
    protected method _return_action {}    
    protected method _browse_action {w}; ## \type TclWidget
    protected method _keypress_action {w k}; ## \type TclWidget, TclKeyBinding
    protected method _motion_action {w x y }; ## \type TclWidget,TclNumber,TclNumber
    protected method _setCount {n}; ## \type TclNumber

    ## abstract method called in refreshid
    protected method refreshrow {row item}; ## \type TclNumber, TclString

    # private methods
    
    private method _getMatchingRows {regex varnameListIDs varnameMapIDs}; ## \type TclString,TclString,TclString
    private method _evalSelectionPatterns {}; ## returns list of matching IDs 

    constructor {args} {}; ## \type TclList

    # protected data members

    ## string of current selected row ids
    protected variable _selection ""
    
    ## column index of table that contains selectable ids (used in _browse_action)
    protected variable _idColumnIndex ""
    
  # string of letters typed by the user for quick access
  private variable _prefix ""

};

## ----------------------------------------------------------------------------
## A DataBrowser constructor.
## \param args arguments passed to itk_initialize
## ----------------------------------------------------------------------------
body DataBrowser::constructor {args} {
   set _idColumnIndex 1
   itk_component add childsite {
      frame $itk_interior.childsite \
		-borderwidth 0 \
		-relief flat \
   } { 
      keep -cursor -background
   }

   itk_component add entryfield {
      iwidgets::entryfield $itk_interior.entry \
		-textvariable [scope _selection] \
		-labelpos w \
		-textbackground gray80 \
		-labeltext "Selection :"
   } {
      keep -background -cursor -foreground 
      rename -labeltext -entrytext entryText  EntryText
      rename -textfont -entryfont entryFont Font
      rename -textbackground -entrybackground entryBackground Background
   }

   itk_component add table {
      MyTable $itk_interior.table \
		-exportselection 0 \
		-rows 20 \
		-cols 9 \
		-cellanchor c \
		-state disabled \
		-colstretchmode all \
		-colseparator " " \
		-rowstretchmode none \
		-selectmode extended \
		-selecttype row \
		-titlecols 1 \
		-titlerows 1 \
		-rowtagcommand [code $this _rowtag]
   } { 
      keep -background -cursor 
   }

   itk_component add countlabel {
      label $itk_interior.countlabel \
		-relief flat \
		-borderwidth 0 \
		-text "Count: 0"
   } { 
      keep -background -cursor
   }

   itk_component add echo_area {
      label $itk_interior.echo_area \
		-relief flat \
		-borderwidth 0
   } { 
      keep -background -cursor
   }

   # packing
   grid $itk_component(childsite) -row 0 -column 0 -sticky ew -padx 3
   grid $itk_component(countlabel) -row 0 -column 1 -sticky ew -padx 3
   grid $itk_component(entryfield) -row 1 -column 0 -sticky ew -padx 3 -pady 3
   grid $itk_component(echo_area) -row 1 -column 1 -sticky ew -padx 3 -pady 3
   grid $itk_component(table) -row 2 -column 0 -columnspan 2 -sticky news -padx 3
   grid rowconfigure $itk_interior 2 -weight 1
   grid columnconfigure $itk_interior 0 -weight 1

    # some bindings
    set table [$itk_component(table) component table]
    bind $table <Motion> "[code $this _motion_action %W %x %y]; break"
    set ef [$itk_component(entryfield) component entry]
    bind $ef <Return> [code $this _return_action]
    #bind $ef <KeyRelease> [code "$this _keypress_action; continue"]
    bind $table <ButtonRelease-1> [code $this _browse_action %W]
  bind $table <KeyPress> [code "$this _keypress_action %A %K; break"]

   # Default Table Configurations
   $table tag config default
   $table tag config colored -bg gray80

   eval itk_initialize $args
}


## ----------------------------------------------------------------------------
## browse slot.
## This method adjusts the selected file in AllFiles::_selection.
## \param w the widget bound to this slot.
## ----------------------------------------------------------------------------
body DataBrowser::_browse_action {w} {

   set ids ""
   set lastrow 0
   foreach index [$w curselection] {
      scan $index "%d,%d" row col
      if {$lastrow == $row } {
	 continue
      }
      set lastrow $row
      set id [$w get $row,$_idColumnIndex] 
      if {$id != ""} {
	 lappend ids $id
      }
   }
   eval setSelection $ids
}


## ----------------------------------------------------------------------------
## Returns a list of all row indices matching 'regex' in the ID column
## \param regex regular expression string  
## \param varnameListIDs in-out-parameter, name of local variable to
## append matching IDs
## \param varnameMapIDs in-out-parameter, name of local array-variable
## to insert matching IDs
##        in order to prohibit duplicate selection entries in entry field
## ----------------------------------------------------------------------------
body DataBrowser::_getMatchingRows {regex varnameListIDs varnameMapIDs} {
    upvar $varnameListIDs matchingIDs
    upvar $varnameMapIDs mapIDs
    set table $itk_component(table)

    set rmax [$table cget -rows]
    set ids [$table getCell 1,$_idColumnIndex $rmax,$_idColumnIndex]
  
    set rows ""
    foreach id $ids {
	if {[regexp -expanded -- "^$regex$" $id match] == 1} {
	    set row [lsearch $ids $match]
	
	    if {$row >= 0 && [info exists mapIDs($match)] == 0} {
		lappend rows $row
		lappend matchingIDs $match
		set mapIDs($match) 1
	    }
	}
    }
    return $rows
}


## ----------------------------------------------------------------------------
## actions to take place on pressing return in the entryfield.
## ----------------------------------------------------------------------------
body DataBrowser::_return_action {} {
    set _selection [_evalSelectionPatterns]
    set rows [eval setSelection $_selection]
   
   if {$rows != {} } {
      $itk_component(table) see "[lindex $rows end],0"
   }
}

## ----------------------------------------------------------------------------
## Evaluates input to the entry field as a list of regular expressions
## and matches them against all IDs in order to return a list of matching
## IDs 
## ----------------------------------------------------------------------------
body DataBrowser::_evalSelectionPatterns {} {
   set table $itk_component(table)

  
   set rows ""
   set matchingIDs ""
   set mapIDs(0) ""
  
   foreach pattern $_selection {
       set match [_getMatchingRows $pattern matchingIDs mapIDs]
      
       if { $match != "" } {
	   set rows [concat $rows $match] 
       }
     
       if {[llength $rows] > 0} {
	   foreach row $rows {
	       $table selection set $row,0
	   }
       }
   }

   if {$rows != {} } {
      $itk_component(table) see "[lindex $rows end],0"
   }
    
   return $matchingIDs

}


## ----------------------------------------------------------------------------
## return a list of selected ids
## ----------------------------------------------------------------------------
body DataBrowser::getSelection {} {
   return $_selection
}



## ----------------------------------------------------------------------------
## Select one or more rows. A previous selection is cleared;
## without arguemnts, removes all selections.
## 
## ARGS must be a list of strings without spaces in them.
## ----------------------------------------------------------------------------
body DataBrowser::setSelection {args} {
   set table $itk_component(table)

   set rmax [$table cget -rows]
   set ids [$table getCell 0,$_idColumnIndex $rmax,$_idColumnIndex]
   $table selection clear 0,0 end
  
   set newselection ""
   set rows ""
   foreach id $args {
      set row [lsearch $ids $id]
      if {$row >= 0} {
	 $table selection set $row,0
	 lappend newselection $id
	 lappend rows $row
      }
   }
   set _selection $newselection

   return $rows
}

## ----------------------------------------------------------------------------
## Select one or more rows. A previous selection is cleared;
## without arguemnts, removes all selections.
## 
## ARGS is a list of row indices.
## ----------------------------------------------------------------------------
body DataBrowser::setIndexedSelection {args} {

  set table $itk_component(table)
  set rmax [$table cget -rows]
  set ids [$table getCell 0,$_idColumnIndex $rmax,$_idColumnIndex]
  $table selection clear 0,0 end

  set newselection ""
  set rows ""
  foreach row $args {

    if {$row >= 0 && $row <= $rmax} {
      $table selection set $row,0
      set id [$table getCell $row,$_idColumnIndex]
      lappend newselection $id
    }

    $table yview [expr $row - 1]

  }
  set _selection $newselection
}


## ----------------------------------------------------------------------------
## colorize the table rows
## This method is a callback configured to the table in order to colorize the
## rows.
## ----------------------------------------------------------------------------
body DataBrowser::_rowtag {row} {
  if {[expr $row % 2]} {
    return "default"
  } else {
    return "colored"
  }
}


## ----------------------------------------------------------------------------
## display the count-label.
## This number should reflect the number of items selected
## \param n the number to be set
## ----------------------------------------------------------------------------
body DataBrowser::_setCount {n} {
  $itk_component(countlabel) configure -text "Count: $n"
  if {$n < 20} {
    set n 20
  }
  $itk_component(table) configure -rows [expr $n + 1]
}


## ----------------------------------------------------------------------------
## Default motion slot.
## \param w the widget where the motion was detected
## \param x the x coords of the mouse 
## \param y the y coords of the mouse
## ----------------------------------------------------------------------------
body DataBrowser::_motion_action {w x y} {
    focus $w
}


## ----------------------------------------------------------------------------
## refresh the displayed data for a specific ID
## ----------------------------------------------------------------------------
body DataBrowser::refreshid {id} {
  set data [getCData $id]
  if {$data == "NULL"} return

  set table $itk_component(table)
  set rmax [$table cget -rows]
  set ids [$table getCell 0,0 $rmax,0]
  set r [lsearch -exact $ids $id]

  if {$r == -1} {
    init_data
  } else {
    refreshrow $r $data
  }
}


## ----------------------------------------------------------------------------
## React to a keypress into the table.
##
## Keys typed by the user are collected into a string, and the row
## whose id matches that string is selected.
## ----------------------------------------------------------------------------
body DataBrowser::_keypress_action {a k} {

  if {"Next" == $k} {
    scan [$itk_component(table) yview] "%f %f" a z
    set a0 [expr $a + ($z - $a)]
    $itk_component(table) yview moveto $a0
    return
  }

  if {"Prior" == $k} {
    scan [$itk_component(table) yview] "%f %f" a z
    set a0 [expr $a - ($z - $a)]
    $itk_component(table) yview moveto $a0
    return
  }

  if {"Home" == $k} {
    $itk_component(table) yview moveto 0
    return
  }

  if {"End" == $k} {
    $itk_component(table) yview moveto 1
    return
  }

  if {"Down" == $k} {
    scan [$itk_component(table) yview] "%f %f" a z
    if {1.0 == $z} return
    set nrrows [$itk_component(table) cget -rows]
    set offset [expr 1.0 / $nrrows]

    # In very large tables, the offset that we have calculated may
    # be lost in rounding. Therefore we scroll until the yview changes.
    set newa $a
    set n 0
    while {$newa == $a} {
      incr n
      set a0 [expr $a + $n * $offset]
      $itk_component(table) yview moveto $a0
      scan [$itk_component(table) yview] "%f %f" newa z
    }
    return
  }

  if {"Up" == $k} {
    scan [$itk_component(table) yview] "%f %f" a z
    if {0 == $a} return
    set nrrows [$itk_component(table) cget -rows]
    set offset [expr 1.0 / $nrrows]

    # In very large tables, the offset that we have calculated may
    # be lost in rounding. Therefore we  scroll until the yview changes.
    set newa $a
    set n 0
    while {$newa == $a} {
      incr n
      set a0 [expr $a - $n * $offset]
      $itk_component(table) yview moveto $a0
      scan [$itk_component(table) yview] "%f %f" newa z
    }
    return
  }


  # ignore other unprintables such as Shift_left
  if {"" == $a} return

  set table $itk_component(table)
  set rmax [$table cget -rows]
  set ids [$table getCell 1,$_idColumnIndex $rmax,$_idColumnIndex]

  if {"BackSpace" == $k} {

    # take back one character
    regsub ".$" $_prefix "" _prefix
  } elseif {"Escape" == $k} {

    # cancel search prefix
    set _prefix ""
    $itk_component(echo_area) configure -text ""
    setSelection
    return
  } elseif {"Tab" == $k} {

    # intelligently add characters

    # find all keys that start like this
    set next ""
    foreach id $ids {
      if {$id == $_prefix} return
      if {[regexp "^${_prefix}(.)" $id dummy c] != 1} continue
      if {"" == $next} { 
        set next $c
      }
      if {"" != $next && $next != $c} {
        return
      }
    }
    
    # if they all continue the same, add that key to _prefix and repeat
    if {"" != $next} {
      append _prefix $next
      $itk_component(echo_area) configure -text $_prefix
      _keypress_action $a $k
    }

  } else {
    append _prefix $a
  }

  # find first matching row
  set found ""
  set row 0
  foreach id $ids {
    incr row
    if {[regexp "^$_prefix" $id] == 1} {
      set found $id
      break
    }
  }
  if {"" == $found} {
    set id ""
    set row -1
  }

  # redisplay
  $itk_component(echo_area) configure -text $_prefix
  if { -1 != $row} {
    eval setIndexedSelection $row
  }
}

