# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## VisFile - visualisation of a file or directory.
## This class is mainly used by VisDirectory.
##
## \author Michael Daum
##
## $Id: VisFile.tcl,v 1.9 2004/02/25 14:42:09 micha Exp $
## ----------------------------------------------------------------------------
class VisFile {
  
  # -- variables -------------------------------------------------------------

  ## name of the file
  public variable name ""

  ## icon used to represent this file in a VisDirectory
  public variable icon ""

  ## file type
  public variable type ""

  ## link to the VisDirectory that displays this file.
  private variable _visDirectory ""

  ## x position where this file is drawn
  private variable _xCoord -100

  ## y position where this file is drawn
  private variable _yCoord -100

  ## flag indicating whether this file is being selected in a VisDirectory
  private variable _isSelected 0

  ## icon item tag in the drawing canvas
  private variable _iconTag ""

  ## text item tag in the drawing canvas
  private variable _textTag ""

  ## rectangular item tag in the drawing canvas
  private variable _recTag ""

  # -- methods ---------------------------------------------------------------
  public method show {}
  public method hide {}
  public method setCoords {x y}; ## \type TclNumber, TclNumber
  public method getCoords {}
  public method getBBox {}
  public method getWidth {}
  public method getHeight {}
  public method setSelection {value}; ## \type TclNumber
  public method getSelection {}

  constructor {visDir args} {}; ## \type VisDirectory, TclList
  destructor { hide }
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body VisFile::constructor {visDir args} {
  set _visDirectory $visDir
  eval configure $args

  show
}

## ----------------------------------------------------------------------------
## show
## ----------------------------------------------------------------------------
body VisFile::show {} {

  if {$_recTag == ""} {
    set canvas [$_visDirectory component canvas]
    set background [$canvas cget -background]
    set _recTag \
      [$_visDirectory create rectangle 0 0 0 0 \
	-fill $background \
	-outline "" \
	-width 0 \
	-tags "$this rec"]
  }
  
  if {$_iconTag == ""} {
    set _iconTag \
      [$_visDirectory create image  0 0 \
	-image $icon \
	-anchor nw \
	-tags "$this icon"]
  }

  if {$_textTag == ""} {
    set _textTag \
      [$_visDirectory create text 0 0 \
	-font {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*} \
	-anchor nw \
	-text "$name" \
	-tags "$this text"]
  } 

  setCoords $_xCoord $_yCoord
}

## ----------------------------------------------------------------------------
## hide
## ----------------------------------------------------------------------------
body VisFile::hide {} {

  if {$_iconTag != ""} {
    $_visDirectory delete $_iconTag
    set _iconTag ""
  }
  
  if {$_textTag != ""} {
    $_visDirectory delete $_textTag
    set _textTag ""
  }

  if {$_recTag != ""} {
    $_visDirectory delete $_recTag
    set _recTag ""
  }
  
}

## ----------------------------------------------------------------------------
## setCoords
## ----------------------------------------------------------------------------
body VisFile::setCoords {x y} {
  set _xCoord $x
  set _yCoord $y

  if {$_recTag != ""} {
    set maxWidth [$_visDirectory getMaxFileWidth]
    set maxHeight [$_visDirectory getMaxFileHeight]
    set x2 [expr $x + $maxWidth]
    set y2 [expr $y + $maxHeight]
    $_visDirectory coords $_recTag $x $y $x2 $y2
  }

  if {$_iconTag != ""} {
    $_visDirectory coords $_iconTag $x $y
    set iconBBox [$_visDirectory bbox $_iconTag]
    scan $iconBBox "%f %f %f %f" x1 y1 x2 y2
    set x [expr $x + $x2 - $x1 + 4]
  } 
 
  if {$_textTag != ""} {
    $_visDirectory coords $_textTag $x $y
  }

}

## ----------------------------------------------------------------------------
## getCoords
## ----------------------------------------------------------------------------
body VisFile::getCoords {} {
  return "$_xCoord $_yCoord"
}

## ----------------------------------------------------------------------------
## getBBox
## ----------------------------------------------------------------------------
body VisFile::getBBox {} {
  return [eval $_visDirectory bbox $_iconTag $_textTag]
}

## ----------------------------------------------------------------------------
## getWidth
## ----------------------------------------------------------------------------
body VisFile::getWidth {} {
  set bbox [getBBox]
  set x1 [lindex $bbox 0]
  set x2 [lindex $bbox 2]
  return [expr $x2 - $x1]
}

## ----------------------------------------------------------------------------
## getHeight
## ----------------------------------------------------------------------------
body VisFile::getHeight {} {
  set bbox [getBBox]
  set y1 [lindex $bbox 1]
  set y2 [lindex $bbox 3]
  return [expr $y2 - $y1]
}

## ----------------------------------------------------------------------------
## set the selection state.
## This method changes the _isSelected state and configures the item accordingly in
## the _visDirectory.
## \param value possible values are 
## - 1: marks the VisFile in the _visDirectory
## - 0: unmarks the VisFile
## ----------------------------------------------------------------------------
body VisFile::setSelection {value} {
  if {$value} {
    $_visDirectory itemconfigure $_recTag \
      -fill blue4
    $_visDirectory itemconfigure $_textTag \
      -fill white
  } else {
    set canvas [$_visDirectory component canvas]
    set background [$canvas cget -background]

    $_visDirectory itemconfigure $_recTag \
      -fill $background
    $_visDirectory itemconfigure $_textTag \
      -fill black
  }

  set _isSelected $value
}

## ----------------------------------------------------------------------------
## getSelection
## ----------------------------------------------------------------------------
body VisFile::getSelection {} {
  return $_isSelected
}
