# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaGrammar - bundled information about a CDG grammar
## \ingroup YadaConfiguration
##
## \author Michael Daum
##
## $Id: YadaGrammar.tcl,v 1.24 2004/10/15 17:24:37 micha Exp $
## ----------------------------------------------------------------------------
class YadaGrammar {
  inherit YadaConfigItem

  # variables ----------------------------------------------------------------
  
  ## list of all files making up the grammar
  public variable files {}

  ## list of selected wordgraph ids
  private variable _selection ""

  ## array mapping wordgraph ids to YadaWordgraph s
  private variable _wordgraphs; ## \type TclArray

  ## flag indicating whether the grammars _wordgraphs are up to date. If not
  ## we reconstruct this array in scan() or rescan().
  private variable _isUpToDate 0

  ## array mapping selection attribute to wordgraph-id
  private variable _selected 

  # methods ------------------------------------------------------------------

  constructor {args} {}; ## \type TclList
  destructor {}

  public method deleteWordgraphs {}
  public method getWordgraphs {}
  public method getWordgraph {wgName}; ## \type TclString
  public method setWordgraph {wordgraph}; ## \type YadaWordgraph
  public method setSelection {args}; ## \type TclList
  public method getSelection {}
  public method scan {}
  public method rescan {}
  public method clone {args}; ## \type TclList
  public method toDOM {}
  public method fromDOM {rootNode}; ## \type domNode
  public method isUpToDate {}
  public method persistenceFileInSync {}
  public method getPersistanceFileName {}; ## overriding virtual method
 
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaGrammar::constructor {args} {
  global env

  eval configure $args
}

## ----------------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------------
body YadaGrammar::destructor {} {
  deleteWordgraphs
}


## ----------------------------------------------------------------------------
## Serializes Object into a DOM-XML-Node
## @return root DOM node representing the Grammar-Object (root tag <yada>)
## ----------------------------------------------------------------------------
body YadaGrammar::toDOM {} {
    set document [dom createDocument "yada"]
    set root [$document documentElement]

    dom createNodeCmd elementNode grammar
    dom createNodeCmd elementNode file

    $root appendFromScript {
	grammar -name $name  {   
	    foreach f $files {
		file -name $f
	    }
	}
    }

    set nodeConfigDoc [$root selectNodes "/yada/grammar"]

    foreach wordgraphID $_selection {
	set wordgraph $_wordgraphs($wordgraphID)
	$wordgraph configure -isSelected 1
	$nodeConfigDoc appendChild [$wordgraph toDOM]

    }
    foreach wordgraph [getWordgraphs] {
	if {[info exists _selected([$wordgraph cget -name])]==0 || $_selected([$wordgraph cget -name])==0 } {
	    $wordgraph configure -isSelected 0
	    $nodeConfigDoc appendChild [$wordgraph toDOM]
	}
    }
    return $root
}


## ----------------------------------------------------------------------------
## Initialize Object with DOM node (that has saved the object state)
## @param rootNode root DOM-Node (Tag <yada>)
## ----------------------------------------------------------------------------
body YadaGrammar::fromDOM {rootNode} {
    set xpath "/yada/grammar"
    set nodeGrammar [$rootNode selectNodes $xpath]
    set name [$nodeGrammar getAttribute name]

    set nodeFiles [$rootNode selectNodes ${xpath}/file]
    set files {}
    foreach nodeFile $nodeFiles {
	lappend files [$nodeFile getAttribute "name"]
    }

   

    set newSelection ""
    deleteWordgraphs
    set nodeWordgraphs [$rootNode selectNodes ${xpath}/yadaWordgraph]
    foreach nodeWordgraph $nodeWordgraphs {
	set newWordgraph [YadaWordgraph ::#auto]
	$newWordgraph fromDOM $nodeWordgraph
        set _wordgraphs([$newWordgraph cget -name]) $newWordgraph 
	if {[$newWordgraph cget -isSelected]} {
	    lappend newSelection [$newWordgraph cget -name] 
	}
    }
    eval setSelection $newSelection
  
}

## ----------------------------------------------------------------------------
## deleteWordgraphs
## ----------------------------------------------------------------------------
body YadaGrammar::deleteWordgraphs {} {
  foreach wgName [array names _wordgraphs] {
    itcl::delete object $_wordgraphs($wgName)
    unset _wordgraphs($wgName)
  }
  setModified
}

## ----------------------------------------------------------------------------
## getWordgraphs
## ----------------------------------------------------------------------------
body YadaGrammar::getWordgraphs {} {
  set result ""
  foreach wgName [lsort -dictionary [array names _wordgraphs]] {
    lappend result $_wordgraphs($wgName)
  }

  return "$result"
}

## ----------------------------------------------------------------------------
## getWordgraph
## ----------------------------------------------------------------------------
body YadaGrammar::getWordgraph {wgName} {
  if {[info exists _wordgraphs($wgName)]} {
    return $_wordgraphs($wgName)
  } else {
    return ""
  }
}

## ----------------------------------------------------------------------------
## setWordgraph
## registers a YadaWordgraph object
## ----------------------------------------------------------------------------
body YadaGrammar::setWordgraph {wordgraph} {
  set wgName [$wordgraph cget -name]
  if {[info exists _wordgraphs($wgName)]} {
    itcl::delete object $_wordgraphs($wgName)
  }
  set _wordgraphs($wgName) $wordgraph
}

## ----------------------------------------------------------------------------
## scan
## ----------------------------------------------------------------------------
body YadaGrammar::scan {} {
  if {!$_isUpToDate} {
    rescan
  }
}

## ----------------------------------------------------------------------------
## rescan
## ----------------------------------------------------------------------------
body YadaGrammar::rescan {} {
  global env

  .main printStatus "... rescanning grammar '$name'"

  setModified

  # remove all contained wordgraphs from the grammar
  deleteWordgraphs


  # build a temporary xml log file
  set xmlFile [::file join $env(YADA_TMP) _yada[pid].xml.gz]


  # build a cdg process which is to be asked about wordgraphs and annotations
  set machine [.main getMachine "localhost"]

  set process [CdgProcess ::#auto \
    -printCommand "" \
    -xmlFile $xmlFile \
    -initFile none]

  # start a cdgp process
  set thisPid [$process start]
  if {$thisPid <= 0} {
    .main printStatus ""
    printMessage "ERROR: cannot start cdgp process $env(YADA_CDGP)"
    return
  }

  # load the grammar files
  foreach file $files {
    if {[regexp {.*\.gz} $file]} {
      $process tell "set preprocessor /usr/bin/zcat"
      $process tell "set prepsuffix .gz"
    } elseif {[regexp {.*\.m4} $file]} {
      $process tell "set preprocessor /usr/bin/m4 -s -P"
      $process tell "set prepsuffix .m4"
    }
    $process tell "load $file"
  }

  # give us the wordgraphs
  $process tell "wordgraph\nannotation\nquit\n"
  .main printStatus "... started cdgp process"
  $process wait

  #itcl::delete object $process

  # get the document data
  .main printStatus "... opening source file '$xmlFile'"
  set fp [open "|zcat $xmlFile" r]
  .main printStatus "... reading source file '$xmlFile'"
  set data [read $fp]
  close $fp

  ### for debugging
  #file delete $xmlFile

  # parse it
  .main printStatus "... parsing source file '$xmlFile'"
  set document [dom parse $data]

  set node [$document documentElement]
  if {$node == ""} {
    printMessage "ERROR: no cdgp logfile while recaning grammar." error
  }

  # parse all lattice nodes
  foreach latticeNode [$node getElementsByTagName "lattice"] {
    
    # build a new yada wordgraph
    set yadaWordgraph [YadaWordgraph ::#auto]
    $yadaWordgraph readXmlNode $latticeNode

    # assert it into the grammar
    .main printStatus "... adding wordgraph '[$yadaWordgraph cget -name]'"
    setWordgraph $yadaWordgraph
  }
  
  # parse all annotations
  foreach annoNode [$node getElementsByTagName "annotation"] {
    set latticeId [$annoNode getAttribute "lattice"]
    set annoId [$annoNode getAttribute "id"]
    set yadaWordgraph [getWordgraph $latticeId]
    if {$yadaWordgraph == ""} {
      .main printStatus "WARNING: no wordgraph '$latticeId' for annotation '$annoId'"
    } else {
      .main printStatus "... adding annotation '$annoId' to wordgraph '$latticeId'"
      $yadaWordgraph readXmlAnno $annoNode
    }
  }

  # delete it
  $document delete
  set document ""
  set _isUpToDate 1
  .main printStatus "done rescanning"

  # reestablish the selection
  eval setSelection $_selection
}

## ----------------------------------------------------------------------------
## setSelection
## ----------------------------------------------------------------------------
body YadaGrammar::setSelection {args} {
    
    foreach wordgraphID $_selection {
	set _selected($wordgraphID) 0
    }
    foreach wordgraphID $args {
	set _selected($wordgraphID) 1
    }

    set _selection $args
    setModified
}

## ----------------------------------------------------------------------------
## getSelection
## ----------------------------------------------------------------------------
body YadaGrammar::getSelection {} {
  return "$_selection"
}

## ----------------------------------------------------------------------------
## clone
## ----------------------------------------------------------------------------
body YadaGrammar::clone {args} {

  # build a new grammar
  set newGrammar [YadaGrammar ::#auto -name $name -files $files]

  # clone all wordgraphs
  foreach wgName [array names _wordgraphs] {
    set newWordgraph [$_wordgraphs($wgName) clone]
    $newGrammar setWordgraph $newWordgraph
  }

  # clone the selection
  eval $newGrammar setSelection $_selection

  eval $newGrammar configure $args
 
  return $newGrammar
}

## ----------------------------------------------------------------------------
## files callback
## ----------------------------------------------------------------------------
configbody YadaGrammar::files {
 
  if {$files == "<none>"} {
    return
  }
  
  set thisFiles $files
  set files ""
  foreach file $thisFiles {
    catch {lappend files [glob $file]}
  }
  setModified
}

## ----------------------------------------------------------------------------
## Returns the value of private member variable _isUpToDate
## ----------------------------------------------------------------------------
body YadaGrammar::isUpToDate {} {
    return $_isUpToDate
}



## ----------------------------------------------------------------------------
## Predicate returning 1, if grammar file has been changed since the
## last saving, otherwise 0
## If no persistant state file exists, 1 is returned
## ----------------------------------------------------------------------------
body YadaGrammar::persistenceFileInSync {} {
    set zippedFileName [getPersistanceFileName].gz 

    if { ![hasSavedState]} {
	return 1
    }

    ::file stat $zippedFileName persistantStateFileAttribs
    set ctime1 $persistantStateFileAttribs(ctime)

    foreach grammarFile $files {
	::file stat $grammarFile grammarFileAttribs
	set ctime2 $grammarFileAttribs(ctime)
	if { [expr {$ctime2 > $ctime1}] } {
	    return 1
	}
    } 

    return 0
}

## ----------------------------------------------------------------------------
## Getting unique file name from registry
## ----------------------------------------------------------------------------
body YadaGrammar::getPersistanceFileName {} {
    return [.main getPersistanceFileNameForGrammar $name]
}

