# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaPlainDocument - abstract yada document.
## \ingroup YadaDocuments
##
## \author Michael Daum 
##
## $Id: YadaPlainDocument.tcl,v 1.11 2004/05/11 14:30:47 ddreyer Exp $
## ----------------------------------------------------------------------------
class YadaPlainDocument {

  # variables ----------------------------------------------------------------
  
  ## name of the document
  public variable name ""
  
  ## flag indicating whether this object functions as combined model-view object (value "true"),
  ## or as a pure model object (value "false": has to be set explicitly when instantiating object)
  public variable hasView "true"
  
  ## flag indicating initialization.
  ## Possible falues are 0 and 1. init() sets this to 1 when it is called
  ## in the init chain calls of inheriting classes, 
  ## i.e. YadaConfigDocument::init() or YadaRunnableDocument::init().
  protected variable _isInitialized 0

  protected variable _modified 0

  ## Denotes time of most recent modification of document's XML file
  ## CAUSED by this object. This allows recognizing cases where a second
  ## object modifies the same file concurrently, in which case the need
  ## may arise to synchronize the object with the new file contents.
  protected variable _lastModificationTime 0


  # methods ------------------------------------------------------------------
  public method load {}
  public method save {}
  public method toDOM {}; ## virtual
  public method fromDOM {node}; ## virtual \type domNode
  public method activationHandle {}; ## \virtual
  public method deActivationHandle {}; ## \virtual
  public method init {} ;## \virtual
  public method setToggle {name {value ""}}; ## \type TclString, TclRef, \virtual

  public method hasSavedState {}
  public method isModified {}
  public method setModified {}
  public method resetModified {}
  public method displayTitle {}
  public method getPersistanceFileName {} ; ## virtual
  public method lastChanged {} 
  
  protected method updateModificationTime {}

  constructor {} {}

};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaPlainDocument::constructor {} {

}

## ----------------------------------------------------------------------------
## activationHandle
## ----------------------------------------------------------------------------
body YadaPlainDocument::activationHandle {} {
    displayTitle

}

## ----------------------------------------------------------------------------
## deActivationHandle
## ----------------------------------------------------------------------------
body YadaPlainDocument::deActivationHandle {} {
}

## ----------------------------------------------------------------------------
## init
## ----------------------------------------------------------------------------
body YadaPlainDocument::init {} {
  if {!$_isInitialized} {
    set _isInitialized 1
  }
}

## ----------------------------------------------------------------------------
## setToggle
## ----------------------------------------------------------------------------
body YadaPlainDocument::setToggle {name {value ""}} {
}


## ----------------------------------------------------------------------------
## Loads current document from file (file name depends on document name)
## ----------------------------------------------------------------------------
body YadaPlainDocument::load {} {
    set _lastUpdate 0 ; ## reset in case there is some value left from last load operation

    set oldCursors [setBusy]

    set zippedFileName [getPersistanceFileName].gz
    if { ![file exists $zippedFileName]} {  
	resetBusy $oldCursors
	error "File $zippedFileName not found."
    }
 
    set fileID [open "|zcat $zippedFileName" {RDONLY}]
    set xmlData [read  $fileID]
    close $fileID
    set document [dom parse $xmlData]
    resetModified
    ## Modified state may be changed by overridden fromDOM methods 
    fromDOM [$document documentElement]  

    resetBusy $oldCursors
    displayTitle
    updateModificationTime
}

## ----------------------------------------------------------------------------
## Saves current document to file (file name depends on document name)
## ----------------------------------------------------------------------------
body YadaPlainDocument::save {} {
    set oldCursors [setBusy]
    set domNode [toDOM]	
    set fileNameToSave [getPersistanceFileName]

    if { [file exists ${fileNameToSave}.gz] } {
	file delete ${fileNameToSave}.gz
    }

    set fileID [open $fileNameToSave {CREAT WRONLY}]
    puts $fileID [$domNode asXML]
    close $fileID
 
    if {[catch {exec gzip -f $fileNameToSave} errMsg]} {
	_resetBusy $oldCursors
	error "Saving failed: $errMsg"
    }
    resetModified
    resetBusy $oldCursors
    displayTitle

    updateModificationTime
}


## ----------------------------------------------------------------------------
## Predicate returning 1, if configuration data has changed, 0 otherwise
## ----------------------------------------------------------------------------
body YadaPlainDocument::isModified {} {
    return $_modified
}

## ----------------------------------------------------------------------------
## Marks item as modified
## ----------------------------------------------------------------------------
body YadaPlainDocument::setModified {} {
    set _modified 1
    displayTitle
}

## ----------------------------------------------------------------------------
## Removes modification mark
## ----------------------------------------------------------------------------
body YadaPlainDocument::resetModified {} {
    set _modified 0
    displayTitle
}

## ----------------------------------------------------------------------------
## Predicate returning 1, if file containing persistant state of YadaGrammar
## exists, otherwise 0
## ----------------------------------------------------------------------------
body YadaPlainDocument::hasSavedState {} {
    set zippedFileName [getPersistanceFileName].gz 
    if { [file exists $zippedFileName] } {
	return 1
    } else {
	return 0
    }
}




## ----------------------------------------------------------------------------
## Displays title in main window (reflecting modification state of document)
## ----------------------------------------------------------------------------
body YadaPlainDocument::displayTitle {} {
    if { [isModified] && $name != "<none>" } {
	wm title .  "$name *** changed *** " 
    } else {
	wm title . "$name"
    }
}

## ----------------------------------------------------------------------------
## Returns time of last update to document XML file or an empty string,
## if there is no such file 
## ----------------------------------------------------------------------------
body YadaPlainDocument::lastChanged {} {
    if { [hasSavedState] } {
	set zippedFileName [getPersistanceFileName].gz
	file stat $zippedFileName fileAttribs
	return $fileAttribs(ctime)
    } else {
	return ""
    }

}

## ----------------------------------------------------------------------------
## Synchronizes internal modification attribute with modification time of
## document's XML file
## ----------------------------------------------------------------------------
body YadaPlainDocument::updateModificationTime {} {
    file stat [getPersistanceFileName].gz fileAttribs
    set _lastModificationTime $fileAttribs(ctime)
}

