#!/bin/sh
#\
exec wish "$0" "$@" 

# ----------------------------------------------------------------------------
# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# ----------------------------------------------------------------------------

## ----------------------------------------------------------------------------
## \mainpage The YADA Reference Manual
## \image html yadalogo-white.gif
## $Id: yada.tcl,v 1.19 2004/02/25 14:42:10 micha Exp $
## \section Overview Overview
## YADA is Yet Another Data Analyser for the CDG parsing environment. Its purpose
## is to ease the task of parser evaluation by allowing the user to configure 
## series of experiments where problem instances are computed and the resulting
## data is presented in a pleasent way. The data generated can be exported
## into an external spredsheat application or converted into raw data for gnuplot, or
## what ever one likes.
##
## YADA offers a document oriented interface where \a documents of certain types
## are loadable in order to generate and display the data with them. During
## runtime an arbitrary number of documents is loadeable each of which can 
## visualize different data or even the same data from a different perspectivity.
## Thus YADA is easily extensible by implementing new document types which the
## user then can instanciate. At the time of this writing the following document
## types are implemented:
## - YadaRunner: this is the most frequently used document type for a straight
##               forward experiment setup.
## - YadaGlsStats: this document type allows the user to examin experiments with
##               a specific focus on the GLS parsing flavour displaying additional
##               statistics only meaningfull for GLS
## - YadaDifference: this also is a runnable document type as the YadaRunner
##               where two experiments are run alternatingly in one document. 
##               Results from both experiments might be compared immediately.
## - YadaRanking: this document type allows the user to load the data from more than
##               one experiment per document and therefor visualize complete 
##               experimental series.
##
## In terms of the implementation there are more document types as can be seen in
## the \link YadaDocuments YadaDocuments module documentation \endlink .
##
## \section Reading How to read this document
## To make your way thru this reference document you might start at the 
## modules section, i.e. 
## \link YadaDocuments YadaDocuments \endlink and the
## \link YadaScheduler YadaScheduler \endlink.
## These try to divide the application into logical parts.  Understanding
## the applications architecture might also be cleared by reading \ref yada.tcl and
## YadaMain. Take a glimpse at the 
## class hierarchy afterwards and this reference manual hopefully becomes understandeable.
## 
## \section Requirements Requirements
## YADA is implemented in Tcl/Tk, more precisely in Itcl/Itk, the object oriented extension
## to Tcl/Tk (comment: no more tcl without itcl;), also using the Iwidgets set of megawidges.
## In addition we use the TkTable widget by Jeffrey Hobbs, and the TDom extension by Jochen Loewer.
## YADA is known to run with 
## - tcl-8.3, tk-8.3: http://tcl.sourceforge.net
## - tktable-2.7: http://tktable.sourceforge.net
## - tdom-0.5apha: loewerj@hotmail.com
##
## I know there are already more recent versions of that software but, you know,
## never touch a running system. You might try YADA with new versions of the components.
## Tell me if that was successfull. 
##
## By the way, YADA is developed on linux and is \a not intended to run on MS Windows
## although TclTk is available on MS Windows. Note that I will not try to code tcl
## platform-independent in order to support MS Windows. YADA will be kept in a shape
## that it runs on unices, no MS Windows.
##
## \section Environment Environment Variables
## As you might already have look at the YADA startup script yada.sh there are a few
## shell environment varialbes which you might be interested in:
## - \b YADA_LIB :
## This is the location where the YADA application scripts are located. Obviously, a
## correct value of this variable is obligatory.
## - \b YADA_IMAGES :
## This is the location where the images are put. If the variable is not set, it defaults
## to \c $YADA_LIB/images. Still not a valid location results in a programm abortion.
## - \b YADA_PATH :
## This variable points to the location, where the yada data is stored needed for
## your experiments, that is the cdgp binary, configuration files, logfiles and xml files.
## If this directory does not exist the first time you start YADA, it is created.
## - \b YADA_ETC :
## This variable points to the location where your configuration are, that is where
## \c experiments.tcl, \c grammar.tcl and machines.tcl are. If not set it defaults to
## \c $YADA_PATH/etc. If the directory is not existing it will be created.
## - \b YADA_DATA :
## This is the path where the database of your experiments is put. Use the value of
## this variable in \c experiments.tcl to organize the data hierarchy properly from here on.
## If this variable is not set it defaults to \c $YADA_PATH/data. A non existing 
## directory location is tried to be created.
## - \b YADA_GRAMMARS :
## This is the location where your dedicated CDG grammar copies might be put to. If 
## the variable is not set, it defaults to \c $YADA_PATH/grammars. If the location
## still does not exist, YADA will refuse to start.
## - \b YADA_TMP :
## Here YADA and possibly you, the user might put temporarily files. Could easily be
## of the value /tmp but defaults to \c $YADA_PATH/tmp if not set.
## - \b YADA_CDGP :
## This is the full path to a default cdgp binary.
## - \b YADA_EDITOR :
## This is the command to start your favorit editor, that YADA should adress in case of.
##
## \section BugTracking Bug Tracking 
## See http://nats-wiki.informatik.uni-hamburg.de/twiki/bin/view/Papa/YadaTracking
##
## \section Coding Coding and documenting style
## Ok, what's to say about coding? Most of the time too much and oneself is leaving
## such a guide over the time. Here are my 2cent coding style guidelines I tried
## to follow til now in YADA:
## - all classes are defined in a separate file, no two or more classes in one
##   file.
## - yada classes are named using the \c Yada prefix
## - if a class used in YADA is not named with the prefix \c Yada, then this
##   has the simple reason that these clases are borowed from somewhere else
##   or are useable outside of YADA aswell (example: FileSelector)
## - private class members, i.e. methods, procs and variables start with a _ 
##   (underscore)
## - try to think about the logical entity a new class belongs to and add it to
##   a module (using the grouping commands in doxygen) 
##
## Hm, I'll add more if I come accross something important enuf to be mentioned here.
##
## \section Copyright Copyright
## Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
##
## \verbatim
## YADA is licensed under the GPL.
##  
## This application is free software; as a special exception the author gives
## unlimited permission to copy and/or distribute it, with or without 
## modifications, as long as this notice is preserved.
## 
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
## implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
## \endverbatim
## ----------------------------------------------------------------------------
## 
## ----------------------------------------------------------------------------

# ----------------------------------------------------------------------------
# Module definitions
# ----------------------------------------------------------------------------

## ----------------------------------------------------------------------------
## \defgroup YadaDocuments YadaDocuments - embedded objects
## This module groups together all classes that form the so called YadaDocuments
## In here we'd like to further separate two groups of documents:
## the YadaConfiguration, dealing with configuration and the YadaRunnableDocuments
## which are in a way active as they are used to generate the data they consist
## of. The YadaDocuments left here are either abstract or not in the sence
## active as runnable documents are ment to be. So the YadaRanking for example
## does not spawn YadaJobs on its own but collects all the data that once has
## been produced by YadaRunnableDocuments. Well and the YadaStartUp splash screen
## simply is temporarily embedded and removed at startup time in the same way 
## as the normal documents are embedded into the gui.
## ----------------------------------------------------------------------------

## ----------------------------------------------------------------------------
## \defgroup YadaConfiguration YadaConfiguration - managing configurations
## \ingroup YadaDocuments
## This submodule spans the YadaDocuments that are used for configuring a set of
## YadaConfigItem objects like the YadaExperiment, YadaGrammar or YadaMachine objects. 
## YadaConfigDocument objects are not embedded into the application directly like the
## other YadaDocuments but first into a YadaConfig object.
## The YadaConfigDocument objects themselves are therefore
## not freely instanciateable as the YadaRunnableDocuments as we always need
## access to the objects starting via the YadaConfig gui. 
## YadaConfigDocument contain a set of YadaConfigItem objects. And in the case
## of the CDG grammars we have configurations object inside the YadaGrammar, that
## is the YadaWordgraph objects.
## ----------------------------------------------------------------------------

## ----------------------------------------------------------------------------
## \defgroup YadaRunnableDocuments YadaRunnableDocuments - also called active documents
## \ingroup YadaDocuments
## This group of classes are implementing so called documents that are \a runnable,
## that is with their help YadaJob objects are generated and started on YadaMachine
## objects.
## ----------------------------------------------------------------------------

## ----------------------------------------------------------------------------
## \defgroup YadaScheduler YadaScheduler - the data workhorse
## This module bundles the classes involved to start YadaJob objects on
## problem instances (YadaWordgraph objects), and span the workload on
## a set of machines.

##----------------------------------------------------------------------------
## \file yada.tcl
## The yada startup script.
## From here on the shell environment is examined and the YadaMain object is
## created. Afterwards the application is initialized via YadaMain::init().
##----------------------------------------------------------------------------

# the yada prefix in YADA_PATH
if {![info exists env(YADA_PATH)]} {
  set env(YADA_PATH) "~/yada"
}
if {![file exists $env(YADA_PATH)]} {
  puts "INFO: YADA_PATH not set, creating directory $env(YADA_PATH)"
  file mkdir $env(YADA_PATH)
} 

# the yada lib in YADA_LIB
if {![info exists env(YADA_LIB)]} {
  set env(YADA_LIB) [file join $env(YADA_PATH) lib]
} 
if {![file exists $env(YADA_LIB)]} {
  set env(YADA_LIB) [pwd]
  puts "ERROR: yada installation not found"
  puts "       please set the YADA_LIB environment variable"
  puts "       or link it to [file $env(YADA_PATH) lib]"
  exit 1
}

# the yada images in YADA_IMAGES
if {![info exists env(YADA_IMAGES)]} {
  set env(YADA_IMAGES) [file join $env(YADA_LIB) images]
}
if {![file exists $env(YADA_IMAGES)]} {
  puts "ERROR: can't locate the yada images"
  puts "       please set the YADA_IMAGES environment variable"
  puts "       or link them to [file join $env(YADA_LIB) images"
  exit 1
}

# the yada temp in YADA_TMP
if {![info exists env(YADA_TMP)]} {
  set env(YADA_TMP) [file join $env(YADA_PATH) tmp]
}
if {![file exists $env(YADA_TMP)]} {
  puts "INFO: YADA_TMP not set, creating directory $env(YADA_TMP)"
  file mkdir $env(YADA_TMP)
}

# the yada etc in YADA_ETC
if {![info exists env(YADA_ETC)]} {
  set env(YADA_ETC) [file join $env(YADA_PATH) etc]
}
if {![file exists $env(YADA_ETC)]} {
  puts "INFO: YADA_ETC not set, creating directory $env(YADA_ETC)"
  file mkdir $env(YADA_ETC)
}

# the yada data"
if {![info exists env(YADA_DATA)]} {
  set env(YADA_DATA) [file join $env(YADA_PATH) data]
}
if {![file exists $env(YADA_DATA)]} {
  puts "INFO: YADA_DATA not set, creating directory $env(YADA_DATA)"
  file mkdir $env(YADA_DATA)
}


# the yada editor in YADA_EDITOR
if {![info exists env(YADA_EDITOR)]} {
  set env(YADA_EDITOR) "gvim"
  puts "INFO: YADA_EDITOR not set, using $env(YADA_EDITOR)"
}

# the grammars
if {![info exists env(YADA_GRAMMARS)]} {
  set env(YADA_GRAMMARS) [file join $YADA_PATH grammars]
}
if {![file exists $env(YADA_GRAMMARS)]} {
  puts "ERROR: don't know where your grammars are"
  puts "       please set the YADA_GRAMMARS environment variable"
  puts "       or link them to [file join $YADA_PATH grammars]."
  exit 1
}

# the cdgp default binary
if {![info exists env(YADA_CDGP)]} {
  set env(YADA_CDGP) [file join $YADA_PATH bin cdgp ]
}

#----------------------------------------------------------------------------
# load additional packages
#----------------------------------------------------------------------------
lappend auto_path $env(YADA_LIB)

package require tdom
package require Iwidgets
package require Tktable
set tk_strictMotif 1


#----------------------------------------------------------------------------
# build the application
#----------------------------------------------------------------------------

source [file join $env(YADA_LIB) YadaOptions.tcl]

wm withdraw .
scan [wm maxsize .] "%d %d" maxx maxy
set maxy [expr $maxy - 50]
wm geometry . ${maxx}x${maxy}+0+0
wm minsize . 800 600
wm title . Yada

# the main window
set yadaMain [YadaMain .main -labelfont {-adobe-helvetica-bold-r-normal--12-*-*-*-*-*-*-*} ]
pack $yadaMain -expand 1 -fill both 

# the fileselector
FileSelector .fileSelector \
  -title "File Selector" \
  -width 500 \
  -height 400 \
  -master "." \
  -modality application

# the tooltip balloon
Balloon .balloon \
  -textforeground black \
  -textbackground lightyellow \
  -anchor nw \
  -delay 100 \
  -font {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*} \
  -padx 9 \
  -pady 3 \
  -relief raised \
  -borderwidth 1

source [file join $env(YADA_LIB) bugfixes.tcl]

# if we run on mosix then lock the yada to the homenode
set lockFileName "/proc/[pid]/lock"
if {[file exists $lockFileName]} {
  set lockFile [open $lockFileName w]
  puts -nonewline $lockFile 1
  close $lockFile
}

wm deiconify .
.main init
